import contextlib
import logging
import shutil
import subprocess
from pathlib import Path

import psutil
import pytest

import salt.utils.platform
from salt.modules.gpg import _homedir_fix

log = logging.getLogger(__name__)

gnupglib = pytest.importorskip("gnupg", reason="Needs python-gnupg library")
PYGNUPG_VERSION = tuple(int(x) for x in gnupglib.__version__.split("."))

pytestmark = [
    pytest.mark.skip_if_binaries_missing("gpg", reason="Needs gpg binary"),
]


@pytest.fixture
def gpghome(tmp_path, modules):
    user = modules.config.option("user")
    if salt.utils.platform.is_windows() and "\\" in user:
        # At least in the test suite, this config option is set
        # including the hostname, so split it off
        user = user.split("\\", maxsplit=1)[1]
    user_info = modules.user.info(user)
    root = tmp_path / "gpghome"
    if salt.utils.platform.is_windows():
        modules["file.mkdir"](
            str(root),
            owner=user_info["uid"],
            grant_perms={
                user_info["uid"]: {
                    "perms": "full_control",
                    "applies_to": "this_folder_subfolders_files",
                }
            },
        )
    else:
        modules["file.mkdir"](
            str(root), user=user_info["uid"], group=user_info["gid"], mode="0700"
        )

    try:
        yield root
    finally:
        # Make sure we don't leave any gpg-agents running behind
        gpg_connect_agent = shutil.which("gpg-connect-agent")
        if gpg_connect_agent:
            gnupghome = root / ".gnupg"
            if not gnupghome.is_dir():
                gnupghome = root
            try:
                subprocess.run(
                    [gpg_connect_agent, "killagent", "/bye"],
                    env={"GNUPGHOME": str(gnupghome)},
                    shell=False,
                    check=True,
                    stdout=subprocess.DEVNULL,
                    stderr=subprocess.DEVNULL,
                )
            except subprocess.CalledProcessError:
                # This is likely CentOS 7 or Amazon Linux 2
                pass

        # If the above errored or was not enough, as a last resort, let's check
        # the running processes.
        for proc in psutil.process_iter():
            try:
                if "gpg-agent" in proc.name():
                    for arg in proc.cmdline():
                        if str(root) in arg:
                            proc.terminate()
            except Exception:  # pylint: disable=broad-except
                pass


@pytest.fixture
def gpg(loaders, states, gpghome):
    try:
        yield states.gpg
    finally:
        pass


@pytest.fixture
def key_a_fp():
    return "EF03765F59EE904930C8A781553A82A058C0C795"


@pytest.fixture
def key_a_pub():
    return """\
-----BEGIN PGP PUBLIC KEY BLOCK-----

mI0EY4fxHQEEAJvXEaaw+o/yZCwMOJbt5FQHbVMMDX/0YI8UdzsE5YCC4iKnoC3x
FwFdkevKj3qp+45iBGLLnalfXIcVGXJGACB+tPHgsfHaXSDQPSfmX6jbZ6pHosSm
v1tTixY+NTJzGL7hDLz2sAXTbYmTbXeE9ifWWk6NcIwZivUbhNRBM+KxABEBAAG0
LUtleSBBIChHZW5lcmF0ZWQgYnkgU2FsdFN0YWNrKSA8a2V5YUBleGFtcGxlPojR
BBMBCAA7FiEE7wN2X1nukEkwyKeBVTqCoFjAx5UFAmOH8R0CGy8FCwkIBwICIgIG
FQoJCAsCBBYCAwECHgcCF4AACgkQVTqCoFjAx5XURAQAguOwI+49lG0Kby+Bsyv3
of3GgxvhS1Qa7+ysj088az5GVt0pqVe3SbRVvn/jyC6yZvWuv94KdL3R7hCeEz2/
JakCRJ4wxEsdeASE8t9H/oTqD0I5asMa9EMvn5ICEGeLsTeQb7OYYihTQj7HJLG6
pDEmK8EhJDvV/9o0lnhm/9w=
=Wc0O
-----END PGP PUBLIC KEY BLOCK-----"""


@pytest.fixture
def key_b_fp():
    return "118B4FAB78038CB2DF7B69E20F6C422647465C93"


@pytest.fixture
def key_b_pub():
    return """\
-----BEGIN PGP PUBLIC KEY BLOCK-----

mI0EY4fxNQEEAOgAzbpheJrOq4il5BrMVtP1G1kU94QX2+xLXEgW/wPdE4HD6Zbg
vliIg18v7Na4x8ubWy/7CkXC83EJ8SoSqcCccvuKjIWsm6tfeCidNstNCjewFMUR
7ZOQmAe/I2JAlz2SgNxS3ZDiCZpGkxqE0GZ+1N7Mz2WHImnExG149RVHABEBAAG0
LUtleSBCIChHZW5lcmF0ZWQgYnkgU2FsdFN0YWNrKSA8a2V5YkBleGFtcGxlPojR
BBMBCAA7FiEEEYtPq3gDjLLfe2niD2xCJkdGXJMFAmOH8TUCGy8FCwkIBwICIgIG
FQoJCAsCBBYCAwECHgcCF4AACgkQD2xCJkdGXJNR3AQAk5ZoN+/ViIX3vA/LbXPn
2VE1E7ETTeIGqsb5f98UfjIbYfkNE8+OtnPxnDbSOPWBEOT+XPPjmxnE0a2UNTfn
ECO71/ZUiyC3ZN50IZ0vgzwBH+DeIV6PDAAun5FGx4RI7v6n0CPlrUcWKYe8wY1F
COflOxnEyLVHXnX8wUIzZwo=
=Hq0X
-----END PGP PUBLIC KEY BLOCK-----"""


@pytest.fixture
def key_c_fp():
    return "96F136AC4C92D78DAF33105E35C03186001C6E31"


@pytest.fixture
def key_c_pub():
    return """\
-----BEGIN PGP PUBLIC KEY BLOCK-----

mI0EY4f2GgEEALToT23wZfLGM/JGCV4pWRlIXXqLwwEBSXral92HvsUjC8Vqsh1z
1n0K8/vIpS9OH2Q21emtht4y36rbahy+w6wRc1XXjPQ28Pyd+8v/jSKy/NKW3g+y
ZoB22vj4L35pAu/G6xs9+pKsLHGjMo+LsWZNEZ2Ar06aYA0dbTb0AqYfABEBAAG0
LUtleSBDIChHZW5lcmF0ZWQgYnkgU2FsdFN0YWNrKSA8a2V5Y0BleGFtcGxlPojR
BBMBCAA7FiEElvE2rEyS142vMxBeNcAxhgAcbjEFAmOH9hoCGy8FCwkIBwICIgIG
FQoJCAsCBBYCAwECHgcCF4AACgkQNcAxhgAcbjH2WAP/RtlUfN/novwmxxma6Zom
P6skFnCcRCs0vMU3OnNwuxZt9B+j0sUTu6noGi04Gcbd0eQs7v57DQHcRhNidZU/
8BJv5jD6E2yuzLK9lON+Yhgc6Pg6raA3hBeCY2HuzTEQLAThyV7ihboNILo7FJwo
y9KvnTFP2+oeDX2Z/m4SoWw=
=81Kb
-----END PGP PUBLIC KEY BLOCK-----"""


@pytest.fixture
def key_e_fp():
    return "2401C402776328D78D6B4C5D67D35BC98502D9B9"


# expires 2022-12-01
@pytest.fixture
def key_e_pub():
    return """\
-----BEGIN PGP PUBLIC KEY BLOCK-----

mI0EY4gjEQEEAKYpWezZQWiAUDvAMcMhBkjHGY2fM4MMiXc6+fRbNV4VCL9TtJYE
gjccYVu44DtIYQzMVimrPQ6xepUmFRalezCG0OO4v25Ciwyeg8LX+Tb3kyAYFAxi
qLXAJyr3aZ/539xBak/Vf5xdURIi7WF5qBGQxd87tRDDqyPFnr87JJtFABEBAAG0
LUtleSBFIChHZW5lcmF0ZWQgYnkgU2FsdFN0YWNrKSA8a2V5ZUBleGFtcGxlPojX
BBMBCABBFiEEJAHEAndjKNeNa0xdZ9NbyYUC2bkFAmOIIxECGy8FCQAAZh8FCwkI
BwICIgIGFQoJCAsCBBYCAwECHgcCF4AACgkQZ9NbyYUC2bmn1QP/WPVhj1bC9/9R
hifv29MG9maRNIkuEkZKtRJj7HMSaamD5IOtGoyMuBwicb38n2Z2KQZUiJbvyZTt
PS328F8YSUSyWQKqmhwL0iLlnDzx8l/nFr5tiss2b/ZzjlMP4iXtAgEdVMJnfjrM
J7xvL0cNSsHha4hUIrekvzM+SNwYkzs=
=nue4
-----END PGP PUBLIC KEY BLOCK-----"""


@pytest.fixture
def key_e_pub_notexpired():
    return """\
-----BEGIN PGP PUBLIC KEY BLOCK-----

mI0EY4gjEQEEAKYpWezZQWiAUDvAMcMhBkjHGY2fM4MMiXc6+fRbNV4VCL9TtJYE
gjccYVu44DtIYQzMVimrPQ6xepUmFRalezCG0OO4v25Ciwyeg8LX+Tb3kyAYFAxi
qLXAJyr3aZ/539xBak/Vf5xdURIi7WF5qBGQxd87tRDDqyPFnr87JJtFABEBAAG0
LUtleSBFIChHZW5lcmF0ZWQgYnkgU2FsdFN0YWNrKSA8a2V5ZUBleGFtcGxlPojR
BBMBCAA7AhsvBQsJCAcCAiICBhUKCQgLAgQWAgMBAh4HAheAFiEEJAHEAndjKNeN
a0xdZ9NbyYUC2bkFAmYOi8gACgkQZ9NbyYUC2bmTyAP+Jo5WUP9LYtXgcbKdhcbz
Kt6Cgbk39rzpmAYpejRSmiu0VrSuSou5W+60YhPPLOVdNOOsKFK1n1wO6sNwCTRU
xrQwNI2yBnuCIV/ZmuOdXLRKc4L8nGXW4lmDKK1PqrXDNH14Bpw0e+FVOR+iR3nW
G5lpc2BZ/RGsECq/HcbpFIM=
=qG1x
-----END PGP PUBLIC KEY BLOCK-----
"""


@pytest.fixture
def gnupg(gpghome):
    _gnupg = gnupglib.GPG(gnupghome=str(gpghome), verbose=True)
    _gnupg.gnupghome = _homedir_fix(gpghome)
    # Force initialization
    _gnupg.list_keys()
    # log gpg binary path, maybe that gives us some hints
    log.error(_gnupg.gpgbinary)
    return _gnupg


@pytest.fixture
def gnupg_keyring(gpghome, keyring):
    _gnupg = gnupglib.GPG(gnupghome=str(gpghome), keyring=keyring)
    _gnupg.gnupghome = _homedir_fix(gpghome)
    return _gnupg


@pytest.fixture(params=["a"])
def _pubkeys_present(gnupg, request, gpghome):
    pubkeys = [request.getfixturevalue(f"key_{x}_pub") for x in request.param]
    fingerprints = [request.getfixturevalue(f"key_{x}_fp") for x in request.param]
    dir_gnupg = dir(gnupg)
    gnupg_version = gnupg.version
    import_result = gnupg.import_keys("\n".join(pubkeys))
    log.error("import result: %r", import_result.__dict__)
    log.error("gpghome stat: %r", gpghome.stat())
    home_contents = list(gpghome.glob("**/*"))
    log.error("gpghome list: %r", home_contents)
    if home_contents:
        log.error("gpg dir file stat: %r", home_contents[1].stat())
    import_count = import_result.count
    import_fingerprints = import_result.fingerprints
    present_keys = gnupg.list_keys()
    for fp in fingerprints:
        assert any(x["fingerprint"] == fp for x in present_keys)
    yield
    # cleanup is taken care of by gpghome and tmp_path


@pytest.fixture(params=["a"])
def keyring(gpghome, tmp_path, request, gpg):
    keyring = tmp_path / "keys.gpg"
    _gnupg_keyring = gnupglib.GPG(gnupghome=str(gpghome), keyring=str(keyring))
    _gnupg_keyring.gnupghome = _homedir_fix(gpghome)

    pubkeys = [request.getfixturevalue(f"key_{x}_pub") for x in request.param]
    fingerprints = [request.getfixturevalue(f"key_{x}_fp") for x in request.param]
    _gnupg_keyring.import_keys("\n".join(pubkeys))
    present_keys = _gnupg_keyring.list_keys()
    for fp in fingerprints:
        assert any(x["fingerprint"] == fp for x in present_keys)
    yield str(keyring)
    # cleanup is taken care of by gpghome and tmp_path


@pytest.fixture(params=(False, True))
def testmode(request):
    return request.param


@pytest.mark.windows_whitelisted
@pytest.mark.usefixtures("_pubkeys_present")
def test_gpg_present_no_changes(gpghome, gpg, gnupg, key_a_fp, testmode):
    assert gnupg.list_keys(keys=key_a_fp)
    ret = gpg.present(
        key_a_fp[-16:],
        trust="unknown",
        gnupghome=str(gpghome),
        keyserver="nonexistent",
        test=testmode,
    )
    assert ret.result
    assert not ret.changes


def test_gpg_present_keyring_no_changes(
    gpghome, gpg, gnupg, gnupg_keyring, keyring, key_a_fp, testmode
):
    """
    The keyring tests are not whitelisted on Windows since they are just
    timing out, possibly because of the two separate GPG instances?
    """
    assert not gnupg.list_keys(keys=key_a_fp)
    assert gnupg_keyring.list_keys(keys=key_a_fp)
    ret = gpg.present(
        key_a_fp[-16:],
        trust="unknown",
        gnupghome=str(gpghome),
        keyserver="nonexistent",
        keyring=keyring,
        test=testmode,
    )
    assert ret.result
    assert not ret.changes


@pytest.mark.windows_whitelisted
@pytest.mark.usefixtures("_pubkeys_present")
def test_gpg_present_trust_change(gpghome, gpg, gnupg, key_a_fp, testmode):
    assert gnupg.list_keys(keys=key_a_fp)
    ret = gpg.present(
        key_a_fp[-16:],
        gnupghome=str(gpghome),
        trust="ultimately",
        keyserver="nonexistent",
        test=testmode,
    )
    assert ret.result is not False
    assert (ret.result is None) is testmode
    assert ret.changes
    assert ret.changes == {key_a_fp[-16:]: {"trust": "ultimately"}}
    key_info = gnupg.list_keys(keys=key_a_fp)
    assert key_info
    assert (key_info[0]["trust"] == "u") is not testmode


def test_gpg_present_keyring_trust_change(
    gpghome, gpg, gnupg, gnupg_keyring, keyring, key_a_fp, testmode
):
    assert not gnupg.list_keys(keys=key_a_fp)
    assert gnupg_keyring.list_keys(keys=key_a_fp)
    ret = gpg.present(
        key_a_fp[-16:],
        gnupghome=str(gpghome),
        trust="ultimately",
        keyserver="nonexistent",
        keyring=keyring,
        test=testmode,
    )
    assert ret.result is not False
    assert (ret.result is None) is testmode
    assert ret.changes
    assert ret.changes == {key_a_fp[-16:]: {"trust": "ultimately"}}
    key_info = gnupg_keyring.list_keys(keys=key_a_fp)
    assert key_info
    assert (key_info[0]["trust"] == "u") is not testmode


# Cannot whitelist source/text tests for Windows since it uses a
# keyring internally, which causes test timeouts for some reason.
def test_gpg_present_source(
    gpghome, gpg, gnupg, key_a_fp, key_a_pub, key_b_pub, key_b_fp, testmode
):
    with pytest.helpers.temp_file(
        "keys", contents=key_a_pub + "\n" + key_b_pub
    ) as keyfile:
        ret = gpg.present(
            key_a_fp[-16:],
            gnupghome=str(gpghome),
            skip_keyserver=True,
            source=str(keyfile),
            test=testmode,
        )
    assert ret.result is not False
    assert (ret.result is None) is testmode
    assert ret.changes
    assert key_a_fp[-16:] in ret.changes
    assert ret.changes[key_a_fp[-16:]]["added"]
    assert bool(gnupg.list_keys(keys=key_a_fp)) is not testmode
    assert not gnupg.list_keys(keys=key_b_fp)


@pytest.mark.parametrize("keyring", ((),), indirect=True)
def test_gpg_present_source_keyring(
    gpghome,
    gpg,
    gnupg,
    gnupg_keyring,
    keyring,
    key_a_fp,
    key_a_pub,
    key_b_pub,
    key_b_fp,
    testmode,
):
    """
    Ensure imports from a list of file sources to a keyring work
    """
    with pytest.helpers.temp_file(
        "keys", contents=key_a_pub + "\n" + key_b_pub
    ) as keyfile:
        ret = gpg.present(
            key_a_fp[-16:],
            gnupghome=str(gpghome),
            skip_keyserver=True,
            source=str(keyfile),
            keyring=keyring,
            test=testmode,
        )
    assert ret.result is not False
    assert (ret.result is None) is testmode
    assert ret.changes
    assert key_a_fp[-16:] in ret.changes
    assert ret.changes[key_a_fp[-16:]]["added"]
    assert not gnupg.list_keys(keys=key_a_fp)
    assert bool(gnupg_keyring.list_keys(keys=key_a_fp)) is not testmode
    assert not gnupg_keyring.list_keys(keys=key_b_fp)


@pytest.mark.skipif(
    PYGNUPG_VERSION < (0, 5, 1), reason="Text requires python-gnupg >=0.5.1"
)
def test_gpg_present_text(
    gpghome, gpg, gnupg, key_a_fp, key_a_pub, key_b_pub, key_b_fp, testmode
):
    concat = key_a_pub + "\n" + key_b_pub
    ret = gpg.present(
        key_a_fp[-16:], gnupghome=str(gpghome), text=concat, test=testmode
    )
    assert ret.result is not False
    assert (ret.result is None) is testmode
    assert ret.changes
    assert key_a_fp[-16:] in ret.changes
    assert ret.changes[key_a_fp[-16:]]["added"]
    assert bool(gnupg.list_keys(keys=key_a_fp)) is not testmode
    assert not gnupg.list_keys(keys=key_b_fp)


@pytest.mark.skipif(
    PYGNUPG_VERSION < (0, 5, 1), reason="Text requires python-gnupg >=0.5.1"
)
@pytest.mark.parametrize("keyring", ((),), indirect=True)
def test_gpg_present_text_keyring(
    gpghome,
    gpg,
    gnupg,
    gnupg_keyring,
    keyring,
    key_a_fp,
    key_a_pub,
    key_b_pub,
    key_b_fp,
    testmode,
):
    """
    Ensure imports from a textual source to a keyring work
    """
    concat = key_a_pub + "\n" + key_b_pub
    ret = gpg.present(
        key_a_fp[-16:],
        gnupghome=str(gpghome),
        keyring=keyring,
        text=concat,
        test=testmode,
    )
    assert ret.result is not False
    assert (ret.result is None) is testmode
    assert ret.changes
    assert key_a_fp[-16:] in ret.changes
    assert ret.changes[key_a_fp[-16:]]["added"]
    assert not gnupg.list_keys(keys=key_a_fp)
    assert bool(gnupg_keyring.list_keys(keys=key_a_fp)) is not testmode
    assert not gnupg_keyring.list_keys(keys=key_b_fp)


@pytest.mark.skipif(
    PYGNUPG_VERSION < (0, 5, 1), reason="Text requires python-gnupg >=0.5.1"
)
def test_gpg_present_text_not_contained(
    gpghome, gpg, gnupg, key_a_fp, key_a_pub, key_b_pub, key_b_fp, key_c_fp
):
    concat = key_a_pub + "\n" + key_b_pub
    ret = gpg.present(key_c_fp[-16:], gnupghome=str(gpghome), text=concat)
    assert not ret.result
    assert not ret.changes
    assert not gnupg.list_keys(keys=key_a_fp)
    assert not gnupg.list_keys(keys=key_b_fp)
    assert "Passed text did not contain the requested key" in ret.comment


def test_gpg_present_multi_source(
    gpghome, gpg, gnupg, key_a_fp, key_a_pub, key_b_pub, key_b_fp
):
    with pytest.helpers.temp_file("keyb", contents=key_b_pub) as keybfile:
        with pytest.helpers.temp_file("keya", contents=key_a_pub) as keyafile:
            ret = gpg.present(
                key_a_fp[-16:],
                gnupghome=str(gpghome),
                skip_keyserver=True,
                source=[str(keybfile), str(keyafile)],
            )
    assert ret.result
    assert ret.changes
    assert key_a_fp[-16:] in ret.changes
    assert ret.changes[key_a_fp[-16:]]["added"]
    assert gnupg.list_keys(keys=key_a_fp)
    assert not gnupg.list_keys(keys=key_b_fp)


def test_gpg_present_source_not_contained(
    gpghome, gpg, gnupg, key_a_fp, key_a_pub, key_b_pub, key_b_fp, key_c_fp
):
    with pytest.helpers.temp_file(
        "keys", contents=key_a_pub + "\n" + key_b_pub
    ) as keyfile:
        ret = gpg.present(
            key_c_fp[-16:],
            gnupghome=str(gpghome),
            skip_keyserver=True,
            source=str(keyfile),
        )
    assert not ret.result
    assert not ret.changes
    assert not gnupg.list_keys(keys=key_a_fp)
    assert not gnupg.list_keys(keys=key_b_fp)
    assert (
        "none of the specified sources were found or contained the (unexpired) key"
        in ret.comment
    )


def test_gpg_present_source_bad_keyfile(
    gpghome, gpg, gnupg, key_a_fp, key_a_pub, key_b_pub, key_b_fp
):
    with pytest.helpers.temp_file(
        "keys", contents=key_a_pub + "\n" + key_b_pub
    ) as keyfile:
        with pytest.helpers.temp_file("badkeys", contents="foobar") as badkeyfile:
            ret = gpg.present(
                key_a_fp[-16:],
                gnupghome=str(gpghome),
                skip_keyserver=True,
                source=["/foo/bar/non/ex/is/tent", str(badkeyfile), str(keyfile)],
            )
    assert ret.result
    assert ret.changes
    assert key_a_fp[-16:] in ret.changes
    assert ret.changes[key_a_fp[-16:]]["added"]
    assert gnupg.list_keys(keys=key_a_fp)
    assert not gnupg.list_keys(keys=key_b_fp)


@pytest.mark.parametrize(
    "method",
    (
        "source",
        pytest.param(
            "text",
            marks=pytest.mark.skipif(
                PYGNUPG_VERSION < (0, 5, 1), reason="Text requires python-gnupg >=0.5.1"
            ),
        ),
    ),
)
def test_gpg_present_import_expired_key(
    method, gpghome, gpg, gnupg, key_e_fp, key_e_pub
):
    """
    Ensure that when a newly imported key is expired, the state fails.
    The key should be imported though if it was not present before.
    """
    if method == "source":
        ctx = pytest.helpers.temp_file("keys", contents=key_e_pub)
    else:
        ctx = contextlib.nullcontext()
        params = {"text": key_e_pub}
    with ctx as inst:
        if method == "source":
            params = {"source": [str(inst)]}
        ret = gpg.present(
            key_e_fp[-16:],
            gnupghome=str(gpghome),
            skip_keyserver=True,
            **params,
        )
    assert ret.result is False
    assert "is expired" in ret.comment
    assert ret.changes
    assert ret.changes[key_e_fp[-16:]]["added"]
    assert gnupg.list_keys(keys=key_e_fp)


@pytest.mark.usefixtures("_pubkeys_present")
@pytest.mark.parametrize("_pubkeys_present", (("e",),), indirect=True)
@pytest.mark.parametrize(
    "method",
    (
        "source",
        pytest.param(
            "text",
            marks=pytest.mark.skipif(
                PYGNUPG_VERSION < (0, 5, 1), reason="Text requires python-gnupg >=0.5.1"
            ),
        ),
    ),
)
def test_gpg_present_expired_key_already_present_fails(
    method, gpghome, gpg, gnupg, key_e_fp, key_e_pub
):
    """
    Ensure that when a present key is expired and no new one can be found,
    the state fails without changes.
    """
    if method == "source":
        ctx = pytest.helpers.temp_file("keys", contents=key_e_pub)
    else:
        ctx = contextlib.nullcontext()
        params = {"text": key_e_pub}
    with ctx as inst:
        if method == "source":
            params = {"source": [str(inst)]}
        ret = gpg.present(
            key_e_fp[-16:],
            gnupghome=str(gpghome),
            skip_keyserver=True,
            **params,
        )
    assert ret.result is False
    if method == "source":
        assert "contained the (unexpired) key" in ret.comment
    else:
        assert "but it's expired" in ret.comment
    assert not ret.changes


@pytest.mark.usefixtures("_pubkeys_present")
@pytest.mark.parametrize("_pubkeys_present", (("e",),), indirect=True)
@pytest.mark.parametrize(
    "method",
    (
        "source",
        pytest.param(
            "text",
            marks=pytest.mark.skipif(
                PYGNUPG_VERSION < (0, 5, 1), reason="Text requires python-gnupg >=0.5.1"
            ),
        ),
    ),
)
def test_gpg_present_expired_key_already_present_refresh(
    method, gpghome, gpg, gnupg, key_e_fp, key_e_pub_notexpired, testmode
):
    """
    Ensure that when a present key is expired and a new one is available,
    the key is reimported.
    """
    if method == "source":
        ctx = pytest.helpers.temp_file("keys", contents=key_e_pub_notexpired)
    else:
        ctx = contextlib.nullcontext()
        params = {"text": key_e_pub_notexpired}
    with ctx as inst:
        if method == "source":
            params = {"source": [str(inst)]}
        ret = gpg.present(
            key_e_fp[-16:],
            gnupghome=str(gpghome),
            skip_keyserver=True,
            **params,
            test=testmode,
        )
    assert ret.result is not False
    assert (ret.result is None) is testmode
    assert "the existing one was expired" in ret.comment
    assert ret.changes
    assert ret.changes[key_e_fp[-16:]]["added"]
    assert ret.changes[key_e_fp[-16:]]["refresh"]


@pytest.mark.usefixtures("_pubkeys_present")
@pytest.mark.parametrize("_pubkeys_present", (("e",),), indirect=True)
def test_gpg_present_expired_key_trust_change(
    gpghome, gpg, gnupg, key_e_fp, key_e_pub_notexpired, testmode
):
    """
    Test that key expiry updates and trust changes work together
    """
    assert gnupg.list_keys(keys=key_e_fp)
    with pytest.helpers.temp_file("keys", contents=key_e_pub_notexpired) as keyfile:
        ret = gpg.present(
            key_e_fp[-16:],
            gnupghome=str(gpghome),
            trust="ultimately",
            skip_keyserver=True,
            source=[str(keyfile)],
            test=testmode,
        )
    assert ret.result is not False
    assert (ret.result is None) is testmode
    assert ret.changes
    assert ret.changes == {
        key_e_fp[-16:]: {"added": True, "refresh": True, "trust": "ultimately"}
    }
    key_info = gnupg.list_keys(keys=key_e_fp)
    assert key_info
    assert (key_info[0]["trust"] == "u") is not testmode


@pytest.mark.windows_whitelisted
def test_gpg_absent_no_changes(gpghome, gpg, gnupg, key_a_fp):
    assert not gnupg.list_keys(keys=key_a_fp)
    ret = gpg.absent(key_a_fp[-16:], gnupghome=str(gpghome))
    assert ret.result
    assert not ret.changes


@pytest.mark.windows_whitelisted
@pytest.mark.usefixtures("_pubkeys_present")
def test_gpg_absent(gpghome, gpg, gnupg, key_a_fp):
    assert gnupg.list_keys(keys=key_a_fp)
    assert not gnupg.list_keys(keys=key_a_fp, secret=True)
    ret = gpg.absent(key_a_fp[-16:], gnupghome=str(gpghome))
    assert ret.result
    assert ret.changes
    assert "deleted" in ret.changes
    assert ret.changes["deleted"]


@pytest.mark.usefixtures("_pubkeys_present")
def test_gpg_absent_from_keyring(gpghome, gpg, gnupg, gnupg_keyring, keyring, key_a_fp):
    assert gnupg.list_keys(keys=key_a_fp)
    assert gnupg_keyring.list_keys(keys=key_a_fp)
    ret = gpg.absent(key_a_fp[-16:], gnupghome=str(gpghome), keyring=keyring)
    assert ret.result
    assert ret.changes
    assert gnupg.list_keys(keys=key_a_fp)
    assert not gnupg_keyring.list_keys(keys=key_a_fp)


@pytest.mark.parametrize("keyring", [""], indirect=True)
def test_gpg_absent_from_keyring_delete_keyring(
    gpghome, gpg, gnupg, gnupg_keyring, keyring, key_a_fp
):
    assert not gnupg_keyring.list_keys()
    assert Path(keyring).exists()
    ret = gpg.absent(
        "abc", gnupghome=str(gpghome), keyring=keyring, keyring_absent_if_empty=True
    )
    assert ret.result
    assert ret.changes
    assert "removed" in ret.changes
    assert ret.changes["removed"] == keyring
    assert not Path(keyring).exists()


@pytest.mark.windows_whitelisted
@pytest.mark.usefixtures("_pubkeys_present")
def test_gpg_absent_test_mode_no_changes(gpghome, gpg, gnupg, key_a_fp):
    assert gnupg.list_keys(keys=key_a_fp)
    ret = gpg.absent(key_a_fp[-16:], gnupghome=str(gpghome), test=True)
    assert ret.result is None
    assert ret.changes
    assert "deleted" in ret.changes
    assert ret.changes["deleted"]
    assert gnupg.list_keys(keys=key_a_fp)



"""
Conformance scenarios for ``salt.utils.batch_state.progress_batch``.

Each :class:`Scenario` is a declarative spec: an initial BatchState
dict plus an ordered sequence of :class:`Step` s.  A Step captures one
call into :func:`progress_batch`: the ``new_returns`` passed in, the
``now`` clock value, the :class:`Action` it should return, and
optional post-call state assertions on specific BatchState fields.

Scenarios live in their own module so they can be consumed by more
than one test suite — the Phase 1 conformance runner here, and (in
Phase 2) the same scenarios replayed against the refactored sync
driver to prove behavioral parity.

Conventions
-----------

* Minion ids are ``m1``, ``m2``, ... in the order they appear in
  ``all_minions``.  Scenarios assume FIFO dispatch (the first minion
  in ``pending`` is the first to go out).
* ``_ret()`` builds a minimal realistic return payload.  Scenarios
  that care about specific return fields (retcode dicts, failed flags)
  pass them explicitly.
* ``covers`` lists ids of existing sync-batch unit tests whose
  behavior this scenario mirrors.  The harness verifies each id
  resolves so renamed sync tests can't silently unmoor a scenario.
"""

import dataclasses

from salt.utils.batch_state import Action


def make_state(
    *,
    minions,
    batch_size,
    batch_wait=0,
    timeout=60,
    gather_job_timeout=10,
    failhard=False,
    fun="test.ping",
    arg=None,
    tgt="*",
    tgt_type="glob",
    driver="master",
    jid="20260101000000000000",
    created=0.0,
    user="root",
):
    """
    Build a fresh BatchState dict for use as a scenario's ``initial``.

    Deliberately verbose: one factory, all fields visible, so a reader
    of a scenario never has to guess what the default state looks like.
    """
    return {
        "jid": jid,
        "all_minions": list(minions),
        "pending": list(minions),
        "active": {},
        "done": {},
        "failed": {},
        "wait": [],
        "batch_size": batch_size,
        "fun": fun,
        "arg": list(arg or []),
        "kwargs": {},
        "tgt": tgt,
        "tgt_type": tgt_type,
        "failhard": failhard,
        "batch_wait": batch_wait,
        "timeout": timeout,
        "gather_job_timeout": gather_job_timeout,
        "last_progress": created,
        "created": created,
        "halted": False,
        "halted_reason": None,
        "driver": driver,
        "ret": "",
        "user": user,
    }


def _ret(value=True, retcode=0, **extra):
    """Build a realistic minion-return payload."""
    return {"ret": value, "retcode": retcode, **extra}


@dataclasses.dataclass(frozen=True)
class Step:
    """
    One call into :func:`salt.utils.batch_state.progress_batch`.

    ``new_returns`` and ``now`` map directly to the function's
    arguments.  ``expected`` is the :class:`Action` the call must
    return.  ``state_after`` is a dict of BatchState field names to
    their expected post-call values — list only the fields that are
    the point of the scenario; unlisted fields are not asserted.
    """

    expected: Action
    now: float = 0.0
    new_returns: dict = dataclasses.field(default_factory=dict)
    state_after: dict = dataclasses.field(default_factory=dict)


@dataclasses.dataclass(frozen=True)
class Scenario:
    """Declarative conformance scenario."""

    name: str
    description: str
    initial: dict
    steps: tuple
    covers: tuple = ()


# ---------------------------------------------------------------------------
# Seed scenarios.  Each one is intentionally small and focused on a single
# behavior.  When progress_batch() lands in Phase 1 step 1 these become its
# executable specification.
# ---------------------------------------------------------------------------

SCENARIOS = (
    Scenario(
        name="initial_dispatch_fills_first_sub_batch",
        description=(
            "First call with no new returns publishes the first batch_size "
            "minions and records their dispatch timestamps in active."
        ),
        initial=make_state(
            minions=["m1", "m2", "m3", "m4"],
            batch_size=2,
        ),
        steps=(
            Step(
                now=1.0,
                expected=Action(
                    publish=["m1", "m2"],
                    finished_minions={},
                    timed_out_minions=[],
                    halted=False,
                    halted_reason=None,
                ),
                state_after={
                    "pending": ["m3", "m4"],
                    "active": {"m1": 1.0, "m2": 1.0},
                    "done": {},
                    "failed": {},
                    "halted": False,
                },
            ),
        ),
        covers=("test_single_jid_across_batch_iterations",),
    ),
    Scenario(
        name="subsequent_dispatch_fills_next_sub_batch",
        description=(
            "When the active slots return, progress_batch moves them to "
            "done and dispatches the next pending minions in the same call."
        ),
        initial=make_state(
            minions=["m1", "m2", "m3", "m4"],
            batch_size=2,
        ),
        steps=(
            Step(
                now=1.0,
                expected=Action(
                    publish=["m1", "m2"],
                    finished_minions={},
                    timed_out_minions=[],
                    halted=False,
                    halted_reason=None,
                ),
            ),
            Step(
                now=2.0,
                new_returns={"m1": _ret(), "m2": _ret()},
                expected=Action(
                    publish=["m3", "m4"],
                    finished_minions={"m1": _ret(), "m2": _ret()},
                    timed_out_minions=[],
                    halted=False,
                    halted_reason=None,
                ),
                state_after={
                    "pending": [],
                    "active": {"m3": 2.0, "m4": 2.0},
                    "done": {"m1": _ret(), "m2": _ret()},
                },
            ),
        ),
        covers=("test_single_jid_across_batch_iterations",),
    ),
    Scenario(
        name="last_partial_sub_batch_dispatches_all_remaining",
        description=(
            "When fewer minions remain than batch_size, progress_batch "
            "dispatches everything that's left in one call."
        ),
        initial=make_state(
            minions=["m1", "m2", "m3"],
            batch_size=2,
        ),
        steps=(
            Step(
                now=1.0,
                expected=Action(
                    publish=["m1", "m2"],
                    finished_minions={},
                    timed_out_minions=[],
                    halted=False,
                    halted_reason=None,
                ),
            ),
            Step(
                now=2.0,
                new_returns={"m1": _ret(), "m2": _ret()},
                expected=Action(
                    publish=["m3"],
                    finished_minions={"m1": _ret(), "m2": _ret()},
                    timed_out_minions=[],
                    halted=False,
                    halted_reason=None,
                ),
                state_after={
                    "pending": [],
                    "active": {"m3": 2.0},
                },
            ),
        ),
    ),
    Scenario(
        name="failhard_on_nonzero_retcode_halts_dispatch",
        description=(
            "With failhard=True, a return whose retcode > 0 sets halted "
            "and prevents further publishes even if slots are available."
        ),
        initial=make_state(
            minions=["m1", "m2", "m3", "m4"],
            batch_size=2,
            failhard=True,
        ),
        steps=(
            Step(
                now=1.0,
                expected=Action(
                    publish=["m1", "m2"],
                    finished_minions={},
                    timed_out_minions=[],
                    halted=False,
                    halted_reason=None,
                ),
            ),
            Step(
                now=2.0,
                new_returns={"m1": _ret(retcode=1)},
                expected=Action(
                    publish=[],
                    finished_minions={"m1": _ret(retcode=1)},
                    timed_out_minions=[],
                    halted=True,
                    halted_reason="failhard",
                ),
                state_after={
                    "pending": ["m3", "m4"],
                    "done": {"m1": _ret(retcode=1)},
                    "halted": True,
                    "halted_reason": "failhard",
                },
            ),
        ),
        covers=("test_single_jid_with_failhard",),
    ),
    Scenario(
        name="retcode_dict_collapses_to_max_for_failhard",
        description=(
            "A return whose retcode is itself a dict (one entry per "
            "sub-call) is evaluated by its maximum value.  Max > 0 plus "
            "failhard triggers halt."
        ),
        initial=make_state(
            minions=["m1", "m2"],
            batch_size=2,
            failhard=True,
        ),
        steps=(
            Step(
                now=1.0,
                expected=Action(
                    publish=["m1", "m2"],
                    finished_minions={},
                    timed_out_minions=[],
                    halted=False,
                    halted_reason=None,
                ),
            ),
            Step(
                now=2.0,
                new_returns={
                    "m1": {
                        "ret": True,
                        "retcode": {"test.ping": 0, "test.retcode": 42},
                    },
                },
                expected=Action(
                    publish=[],
                    finished_minions={
                        "m1": {
                            "ret": True,
                            "retcode": {"test.ping": 0, "test.retcode": 42},
                        },
                    },
                    timed_out_minions=[],
                    halted=True,
                    halted_reason="failhard",
                ),
                state_after={"halted": True, "halted_reason": "failhard"},
            ),
        ),
        covers=("test_run_retcode_dict_takes_max",),
    ),
    Scenario(
        name="batch_wait_blocks_next_dispatch_until_cooldown_expires",
        description=(
            "batch_wait puts a per-slot cooldown stamp in state['wait'] "
            "when a minion returns.  While the stamp is in the future "
            "the slot is not available, so pending minions stay pending."
        ),
        initial=make_state(
            minions=["m1", "m2", "m3", "m4"],
            batch_size=2,
            batch_wait=30,
        ),
        steps=(
            Step(
                now=0.0,
                expected=Action(
                    publish=["m1", "m2"],
                    finished_minions={},
                    timed_out_minions=[],
                    halted=False,
                    halted_reason=None,
                ),
            ),
            Step(
                now=1.0,
                new_returns={"m1": _ret(), "m2": _ret()},
                expected=Action(
                    publish=[],
                    finished_minions={"m1": _ret(), "m2": _ret()},
                    timed_out_minions=[],
                    halted=False,
                    halted_reason=None,
                ),
                state_after={
                    "pending": ["m3", "m4"],
                    "active": {},
                    "done": {"m1": _ret(), "m2": _ret()},
                    "wait": [31.0, 31.0],
                },
            ),
            Step(
                now=15.0,
                expected=Action(
                    publish=[],
                    finished_minions={},
                    timed_out_minions=[],
                    halted=False,
                    halted_reason=None,
                ),
                state_after={"pending": ["m3", "m4"], "wait": [31.0, 31.0]},
            ),
            Step(
                now=32.0,
                expected=Action(
                    publish=["m3", "m4"],
                    finished_minions={},
                    timed_out_minions=[],
                    halted=False,
                    halted_reason=None,
                ),
                state_after={
                    "pending": [],
                    "active": {"m3": 32.0, "m4": 32.0},
                    "wait": [],
                },
            ),
        ),
        covers=("test_run_batch_wait_delays_next_dispatch",),
    ),
    Scenario(
        name="timeout_moves_nonresponsive_minion_to_failed",
        description=(
            "An active minion whose dispatch timestamp is older than "
            "now - (timeout + gather_job_timeout) is moved from active "
            "to failed with reason 'timeout'."
        ),
        initial=make_state(
            minions=["m1", "m2"],
            batch_size=2,
            timeout=60,
            gather_job_timeout=10,
        ),
        steps=(
            Step(
                now=0.0,
                expected=Action(
                    publish=["m1", "m2"],
                    finished_minions={},
                    timed_out_minions=[],
                    halted=False,
                    halted_reason=None,
                ),
            ),
            Step(
                now=100.0,  # 100s elapsed, limit is 70s
                expected=Action(
                    publish=[],
                    finished_minions={},
                    timed_out_minions=["m1", "m2"],
                    halted=False,
                    halted_reason=None,
                ),
                state_after={
                    "active": {},
                    "failed": {"m1": "timeout", "m2": "timeout"},
                },
            ),
        ),
    ),
    Scenario(
        name="normal_completion_leaves_halted_false",
        description=(
            "When the last active minion returns and pending is empty, "
            "halted stays False.  The driver detects completion via "
            "is_batch_done(state), not via action.halted."
        ),
        initial=make_state(
            minions=["m1", "m2"],
            batch_size=2,
        ),
        steps=(
            Step(
                now=1.0,
                expected=Action(
                    publish=["m1", "m2"],
                    finished_minions={},
                    timed_out_minions=[],
                    halted=False,
                    halted_reason=None,
                ),
            ),
            Step(
                now=2.0,
                new_returns={"m1": _ret(), "m2": _ret()},
                expected=Action(
                    publish=[],
                    finished_minions={"m1": _ret(), "m2": _ret()},
                    timed_out_minions=[],
                    halted=False,
                    halted_reason=None,
                ),
                state_after={
                    "pending": [],
                    "active": {},
                    "done": {"m1": _ret(), "m2": _ret()},
                    "halted": False,
                    "halted_reason": None,
                },
            ),
        ),
        covers=("test_single_jid_single_batch",),
    ),
)



"""
    :codeauthor: Jayesh Kariya <jayeshk@saltstack.com>

    Test cases for salt.modules.win_ip
"""

import pytest

import salt.modules.win_ip as win_ip
from salt.exceptions import SaltInvocationError


def test_get_subnet_length():
    """
    Test get subnet length is correct.
    """
    assert win_ip.get_subnet_length("255.255.255.0") == 24
    pytest.raises(SaltInvocationError, win_ip.get_subnet_length, "255.255.0")



from pygments.lexer import bygroups, inherit
from pygments.lexers.configs import TerraformLexer
from pygments.token import Keyword, Name, Punctuation, Whitespace


class VaultPolicyLexer(TerraformLexer):
    aliases = ["vaultpolicy"]
    filenames = ["*.hcl"]
    mimetypes = ["application/x-hcl-policy"]

    tokens = {
        "basic": [
            inherit,
            (
                r"(path)(\s+)(\".*\")(\s+)(\{)",
                bygroups(
                    Keyword.Reserved, Whitespace, Name.Variable, Whitespace, Punctuation
                ),
            ),
        ],
    }


def setup(app):
    app.add_lexer("vaultpolicy", VaultPolicyLexer)
    return {"parallel_read_safe": True}



"""
Unit tests for SSL/TLS certificate identity verification.

These tests verify that certificates contain the correct identity information
(CN and SAN) that can be matched against minion IDs.
"""

import pytest

import salt.utils.files

pytestmark = [
    pytest.mark.core_test,
]


def test_minion_cert_has_id_in_cn(ssl_minion_a_cert_key):
    """
    Test that minion certificate has minion ID in Common Name.

    This is critical for the TLS optimization feature which requires
    matching the certificate identity to the minion ID.
    """
    try:
        from cryptography import x509
        from cryptography.hazmat.backends import default_backend
    except ImportError:
        pytest.skip("cryptography library not available")

    cert_path, _ = ssl_minion_a_cert_key

    # Load certificate
    with salt.utils.files.fopen(cert_path, "rb") as f:
        cert = x509.load_pem_x509_certificate(f.read(), default_backend())

    # Get Common Name from subject
    cn = None
    for attr in cert.subject:
        if attr.oid == x509.oid.NameOID.COMMON_NAME:
            cn = attr.value
            break

    assert cn == "test-minion-a"


def test_minion_cert_has_id_in_san(ssl_minion_a_cert_key):
    """
    Test that minion certificate has minion ID in Subject Alternative Name.

    SAN is the preferred location for hostname/identity verification in modern TLS.
    """
    try:
        from cryptography import x509
        from cryptography.hazmat.backends import default_backend
    except ImportError:
        pytest.skip("cryptography library not available")

    cert_path, _ = ssl_minion_a_cert_key

    # Load certificate
    with salt.utils.files.fopen(cert_path, "rb") as f:
        cert = x509.load_pem_x509_certificate(f.read(), default_backend())

    # Get SAN extension
    san_ext = cert.extensions.get_extension_for_oid(
        x509.oid.ExtensionOID.SUBJECT_ALTERNATIVE_NAME
    )

    # Extract DNS names
    dns_names = [name.value for name in san_ext.value]

    # Verify minion ID is in SAN
    assert "test-minion-a" in dns_names


def test_different_minions_have_different_certs(
    ssl_minion_a_cert_key, ssl_minion_b_cert_key
):
    """
    Test that different minion IDs get different certificates with different identities.

    This ensures that minion A cannot use minion B's certificate to impersonate it.
    """
    try:
        from cryptography import x509
        from cryptography.hazmat.backends import default_backend
    except ImportError:
        pytest.skip("cryptography library not available")

    cert_a_path, _ = ssl_minion_a_cert_key
    cert_b_path, _ = ssl_minion_b_cert_key

    # Load both certificates
    with salt.utils.files.fopen(cert_a_path, "rb") as f:
        cert_a = x509.load_pem_x509_certificate(f.read(), default_backend())
    with salt.utils.files.fopen(cert_b_path, "rb") as f:
        cert_b = x509.load_pem_x509_certificate(f.read(), default_backend())

    # Verify subjects are different
    assert cert_a.subject != cert_b.subject

    # Get CNs
    cn_a = None
    cn_b = None
    for attr in cert_a.subject:
        if attr.oid == x509.oid.NameOID.COMMON_NAME:
            cn_a = attr.value
    for attr in cert_b.subject:
        if attr.oid == x509.oid.NameOID.COMMON_NAME:
            cn_b = attr.value

    assert cn_a == "test-minion-a"
    assert cn_b == "test-minion-b"
    assert cn_a != cn_b


def test_extract_identity_from_cert(ssl_minion_a_cert_key):
    """
    Test a helper function pattern for extracting identity from certificate.

    This demonstrates how the optimization feature should extract and verify
    the minion ID from a peer certificate.
    """
    try:
        from cryptography import x509
        from cryptography.hazmat.backends import default_backend
    except ImportError:
        pytest.skip("cryptography library not available")

    def extract_identity_from_cert(cert):
        """
        Extract identity from a certificate.

        Returns the Common Name and all SAN DNS names.
        """
        # Get CN
        cn = None
        for attr in cert.subject:
            if attr.oid == x509.oid.NameOID.COMMON_NAME:
                cn = attr.value
                break

        # Get SAN DNS names
        san_names = []
        try:
            san_ext = cert.extensions.get_extension_for_oid(
                x509.oid.ExtensionOID.SUBJECT_ALTERNATIVE_NAME
            )
            san_names = [name.value for name in san_ext.value]
        except x509.ExtensionNotFound:
            pass

        return cn, san_names

    cert_path, _ = ssl_minion_a_cert_key

    # Load certificate
    with salt.utils.files.fopen(cert_path, "rb") as f:
        cert = x509.load_pem_x509_certificate(f.read(), default_backend())

    cn, san_names = extract_identity_from_cert(cert)

    assert cn == "test-minion-a"
    assert "test-minion-a" in san_names


def test_verify_identity_matches_minion_id(ssl_minion_a_cert_key):
    """
    Test identity verification logic that should be used in the optimization.

    The TLS optimization should only skip AES encryption when:
    1. TLS is active with valid certs
    2. cert_reqs is CERT_REQUIRED
    3. The peer certificate identity matches the expected minion ID
    """
    try:
        from cryptography import x509
        from cryptography.hazmat.backends import default_backend
    except ImportError:
        pytest.skip("cryptography library not available")

    def verify_cert_identity(cert, expected_id):
        """
        Verify that certificate identity matches expected minion ID.

        Checks both CN and SAN for a match.
        """
        # Check CN
        for attr in cert.subject:
            if attr.oid == x509.oid.NameOID.COMMON_NAME:
                if attr.value == expected_id:
                    return True

        # Check SAN
        try:
            san_ext = cert.extensions.get_extension_for_oid(
                x509.oid.ExtensionOID.SUBJECT_ALTERNATIVE_NAME
            )
            san_names = [name.value for name in san_ext.value]
            if expected_id in san_names:
                return True
        except x509.ExtensionNotFound:
            pass

        return False

    cert_path, _ = ssl_minion_a_cert_key

    # Load certificate
    with salt.utils.files.fopen(cert_path, "rb") as f:
        cert = x509.load_pem_x509_certificate(f.read(), default_backend())

    # Test matching ID
    assert verify_cert_identity(cert, "test-minion-a") is True

    # Test non-matching ID
    assert verify_cert_identity(cert, "test-minion-b") is False
    assert verify_cert_identity(cert, "test-minion-c") is False
    assert verify_cert_identity(cert, "wrong-id") is False


def test_verify_identity_mismatch_prevents_optimization(
    ssl_minion_a_cert_key, ssl_minion_b_cert_key
):
    """
    Test that identity mismatch should prevent the TLS optimization.

    Even if minion A has a valid certificate, if it claims to be minion B,
    the optimization should NOT be used (and AES should still be applied).
    """
    try:
        from cryptography import x509
        from cryptography.hazmat.backends import default_backend
    except ImportError:
        pytest.skip("cryptography library not available")

    def should_skip_aes_with_tls(cert, claimed_minion_id):
        """
        Determine if AES encryption can be skipped.

        Returns True only if certificate identity matches claimed ID.
        """
        # Check CN
        for attr in cert.subject:
            if attr.oid == x509.oid.NameOID.COMMON_NAME:
                if attr.value == claimed_minion_id:
                    return True

        # Check SAN
        try:
            san_ext = cert.extensions.get_extension_for_oid(
                x509.oid.ExtensionOID.SUBJECT_ALTERNATIVE_NAME
            )
            san_names = [name.value for name in san_ext.value]
            if claimed_minion_id in san_names:
                return True
        except x509.ExtensionNotFound:
            pass

        return False

    # Load minion A's certificate
    cert_a_path, _ = ssl_minion_a_cert_key
    with salt.utils.files.fopen(cert_a_path, "rb") as f:
        cert_a = x509.load_pem_x509_certificate(f.read(), default_backend())

    # Minion A claiming to be "test-minion-a" - should allow optimization
    assert should_skip_aes_with_tls(cert_a, "test-minion-a") is True

    # Minion A claiming to be "test-minion-b" - should NOT allow optimization
    assert should_skip_aes_with_tls(cert_a, "test-minion-b") is False

    # Minion A claiming to be something else - should NOT allow optimization
    assert should_skip_aes_with_tls(cert_a, "malicious-minion") is False



import contextlib
import ctypes
import random
from copy import copy

import pytest

import salt.utils.win_lgpo_auditpol as win_lgpo_auditpol
from tests.support.mock import MagicMock, patch

pytestmark = [
    pytest.mark.windows_whitelisted,
    pytest.mark.skip_unless_on_windows,
]


@pytest.fixture(scope="module")
def settings():
    return ["No Auditing", "Success", "Failure", "Success and Failure"]


@pytest.fixture(autouse=True)
def reset_auditpol_api_cache():
    with patch.object(win_lgpo_auditpol, "_API", new=None):
        yield


@pytest.fixture
def configure_loader_modules():
    return {
        win_lgpo_auditpol: {
            "__context__": {},
        }
    }


def test_get_settings():
    names = win_lgpo_auditpol._get_valid_names()
    ret = win_lgpo_auditpol.get_settings(category="All")
    for name in names:
        assert name in [k.lower() for k in ret]


def test_get_settings_invalid_category():
    pytest.raises(KeyError, win_lgpo_auditpol.get_settings, category="Fake Category")


@pytest.mark.slow_test
def test_get_setting(settings):
    names = win_lgpo_auditpol._get_valid_names()
    for name in names:
        ret = win_lgpo_auditpol.get_setting(name)
        assert ret in settings


def test_get_setting_invalid_name():
    pytest.raises(KeyError, win_lgpo_auditpol.get_setting, name="Fake Name")


def test_set_setting(settings):
    def noop_security_privilege():
        return contextlib.nullcontext()

    names = ["Credential Validation", "IPsec Driver", "File System", "SAM"]
    mock_set = MagicMock(return_value=True)
    real_api = win_lgpo_auditpol._load_advapi32()
    patched_api = copy(real_api)
    patched_api.AuditSetSystemPolicy = mock_set
    with patch.object(win_lgpo_auditpol, "_API", patched_api):
        with patch.object(
            win_lgpo_auditpol,
            "_enable_se_security_privilege",
            noop_security_privilege,
        ):
            with patch.object(
                win_lgpo_auditpol,
                "_get_valid_names",
                return_value=[k.lower() for k in names],
            ):
                for name in names:
                    value = random.choice(settings)
                    win_lgpo_auditpol.set_setting(name=name, value=value)
                    mask = win_lgpo_auditpol.settings[value]
                    mock_set.assert_called_once()
                    args, _kwargs = mock_set.call_args
                    assert args[1] == 1
                    policy_ptr = ctypes.cast(
                        args[0],
                        ctypes.POINTER(win_lgpo_auditpol._AUDIT_POLICY_INFORMATION),
                    )
                    assert policy_ptr.contents.AuditingInformation == mask
                    mock_set.reset_mock()


def test_set_setting_invalid_setting():
    names = ["Credential Validation", "IPsec Driver", "File System"]
    with patch.object(
        win_lgpo_auditpol,
        "_get_valid_names",
        return_value=[k.lower() for k in names],
    ):
        pytest.raises(
            KeyError,
            win_lgpo_auditpol.set_setting,
            name="Fake Name",
            value="No Auditing",
        )


def test_set_setting_invalid_value():
    names = ["Credential Validation", "IPsec Driver", "File System"]
    with patch.object(
        win_lgpo_auditpol,
        "_get_valid_names",
        return_value=[k.lower() for k in names],
    ):
        pytest.raises(
            KeyError,
            win_lgpo_auditpol.set_setting,
            name="Credential Validation",
            value="Fake Value",
        )


def test_get_auditpol_dump():
    names = win_lgpo_auditpol._get_valid_names()
    dump = win_lgpo_auditpol.get_auditpol_dump()
    for name in names:
        found = False
        for line in dump:
            if name.lower() in line.lower():
                found = True
                break
        assert found is True


def test_get_advaudit_policy_rows_matches_fieldnames():
    rows = win_lgpo_auditpol.get_advaudit_policy_rows()
    assert rows
    expected = win_lgpo_auditpol._FIELDNAMES
    assert list(rows[0].keys()) == expected



import hashlib
import os
import time
from pathlib import Path

import pytest
from saltfactories.utils import random_string

from tests.support.helpers import system_python_version
from tests.support.runtests import RUNTIME_VARS

pytestmark = [
    pytest.mark.slow_test,
    pytest.mark.skip_on_windows(reason="salt-ssh not available on Windows"),
    pytest.mark.skipif(
        system_python_version() < (3, 10),
        reason="System python too old for these tests",
    ),
]


@pytest.fixture(scope="module", autouse=True)
def _pillar_tree(salt_master):
    top_file = """
    base:
      'localhost':
        - basic
      '127.0.0.1':
        - basic
    """
    basic_pillar_file = """
    test_pillar: cheese
    alot: many
    script: grail
    """
    with salt_master.pillar_tree.base.temp_file(
        "top.sls", top_file
    ), salt_master.pillar_tree.base.temp_file("basic.sls", basic_pillar_file):
        yield


@pytest.fixture(scope="function")
def cachedir(salt_ssh_cli_parameterized):
    """
    The current minion cache dir

    Note: This uses function scope (not module scope) to match salt_ssh_cli_parameterized.
    Module-scoped fixtures cannot depend on function-scoped fixtures in pytest.
    """
    # The salt-ssh cache dir in the minion context is different than
    # the one available in the salt_ssh_cli_parameterized opts. Any other way to get this? TODO
    res = salt_ssh_cli_parameterized.run("cp.cache_dest", "salt://file")
    assert res.returncode == 0
    assert isinstance(res.data, str)
    # This will return <cachedir>/files/base/file
    return Path(res.data).parent.parent.parent


@pytest.fixture(scope="function")
def master_cachedir(salt_ssh_cli_parameterized, cachedir):
    """
    The current per-target master-side staging root for salt-ssh files.

    The salt-ssh ``cp`` wrapper stages master-side files under
    ``<_master_staging_root()>/salt-ssh/<id>/`` -- a *sibling* of the thin
    dir, not a child of the minion cachedir. This is required because
    ``ssh_py_shim`` removes the thin dir tree when ``ssh_wipe`` is enabled,
    which would otherwise wipe staged files before the test (or any caller)
    can inspect them. Discover the actual location at runtime via
    ``cp.convert_cache_path`` so we mirror the production path computation
    in :class:`salt.client.ssh.wrapper.cp.SSHCpClient` instead of
    duplicating its (changing) layout assumptions.
    """
    probe = cachedir / "files" / "base" / "_master_cachedir_probe"
    res = salt_ssh_cli_parameterized.run(
        "cp.convert_cache_path", str(probe), master=True
    )
    assert res.returncode == 0
    assert isinstance(res.data, str)
    # convert_cache_path returns <master_root>/salt-ssh/<id>/files/base/_probe...
    # strip ``files/base/_master_cachedir_probe`` to get the per-target root.
    return Path(res.data).parent.parent.parent


def _convert(cachedir, master_cachedir, path, master=False):
    if not isinstance(path, Path):
        path = Path(path)
    if master:
        if master_cachedir in path.parents:
            return path
        return master_cachedir / path.relative_to(cachedir)
    if master_cachedir not in path.parents:
        return path
    return cachedir / path.relative_to(master_cachedir)


@pytest.mark.parametrize("template", (None, "jinja"))
@pytest.mark.parametrize("dst_is_dir", (False, True))
def test_get_file(
    salt_ssh_cli_parameterized, tmp_path, template, dst_is_dir, master_cachedir
):
    src = "salt://" + ("cheese" if not template else "{{pillar.test_pillar}}")
    if dst_is_dir:
        tgt = tmp_path
    else:
        tgt = tmp_path / ("cheese" if not template else "{{pillar.test_pillar}}")
    res = salt_ssh_cli_parameterized.run(
        "cp.get_file", src, str(tgt), template=template
    )
    assert res.returncode == 0
    assert res.data
    tgt = tmp_path / "cheese"
    assert res.data == str(tgt)
    master_path = master_cachedir / "files" / "base" / "cheese"
    for path in (tgt, master_path):
        assert path.exists()
        data = path.read_text(encoding="utf-8")
        assert "Gromit" in data
        assert "bacon" not in data


def test_get_file_gzipped(salt_ssh_cli_parameterized, caplog, tmp_path):
    tgt = tmp_path / random_string("foo-")
    res = salt_ssh_cli_parameterized.run(
        "cp.get_file", "salt://grail/scene33", str(tgt), gzip=5
    )
    assert res.returncode == 0
    assert res.data
    assert res.data == str(tgt)
    assert "The gzip argument to cp.get_file in salt-ssh is unsupported" in caplog.text
    assert tgt.exists()
    data = tgt.read_text(encoding="utf-8")
    assert "KNIGHT:  They're nervous, sire." in data
    assert "bacon" not in data


def test_get_file_makedirs(salt_ssh_cli_parameterized, tmp_path, master_cachedir):
    tgt = tmp_path / "make" / "dirs" / "scene33"
    res = salt_ssh_cli_parameterized.run(
        "cp.get_file", "salt://grail/scene33", str(tgt), makedirs=True
    )
    assert res.returncode == 0
    assert res.data
    assert res.data == str(tgt)
    master_path = master_cachedir / "files" / "base" / "grail" / "scene33"
    for path in (tgt, master_path):
        assert path.exists()
        data = path.read_text(encoding="utf-8")
        assert "KNIGHT:  They're nervous, sire." in data
        assert "bacon" not in data


@pytest.mark.parametrize("suffix", ("", "?saltenv=prod"))
def test_get_file_from_env(salt_ssh_cli_parameterized, tmp_path, suffix):
    tgt = tmp_path / "cheese"
    ret = salt_ssh_cli_parameterized.run(
        "cp.get_file", "salt://cheese" + suffix, str(tgt)
    )
    assert ret.returncode == 0
    assert ret.data
    assert ret.data == str(tgt)
    data = tgt.read_text(encoding="utf-8")
    assert "Gromit" in data
    assert ("Comte" in data) is bool(suffix)


def test_get_file_nonexistent_source(salt_ssh_cli_parameterized):
    res = salt_ssh_cli_parameterized.run(
        "cp.get_file", "salt://grail/nonexistent_scene", ""
    )
    assert res.returncode == 0  # not a fan of this
    assert res.data == ""


def test_envs(salt_ssh_cli_parameterized):
    ret = salt_ssh_cli_parameterized.run("cp.envs")
    assert ret.returncode == 0
    assert ret.data
    assert isinstance(ret.data, list)
    assert sorted(ret.data) == sorted(["base", "prod"])


def test_get_template(salt_ssh_cli_parameterized, tmp_path, master_cachedir):
    tgt = tmp_path / "scene33"
    res = salt_ssh_cli_parameterized.run(
        "cp.get_template", "salt://grail/scene33", str(tgt), spam="bacon"
    )
    assert res.returncode == 0
    assert res.data
    assert res.data == str(tgt)
    master_path = master_cachedir / "extrn_files" / "base" / "grail" / "scene33"
    for path in (tgt, master_path):
        assert tgt.exists()
        data = tgt.read_text(encoding="utf-8")
        assert "bacon" in data
        assert "spam" not in data


def test_get_template_dest_empty(salt_ssh_cli_parameterized, cachedir, master_cachedir):
    res = salt_ssh_cli_parameterized.run(
        "cp.get_template", "salt://grail/scene33", "", spam="bacon"
    )
    assert res.returncode == 0
    assert res.data
    assert isinstance(res.data, str)
    master_path = master_cachedir / "extrn_files" / "base" / "grail" / "scene33"
    tgt = _convert(cachedir, master_cachedir, master_path)
    assert res.data == str(tgt)
    for file in (tgt, master_path):
        assert file.exists()
        data = file.read_text(encoding="utf-8")
        assert "bacon" in data
        assert "spam" not in data


def test_get_template_nonexistent_source(salt_ssh_cli_parameterized, tmp_path):
    res = salt_ssh_cli_parameterized.run(
        "cp.get_template", "salt://grail/nonexistent_scene", ""
    )
    assert res.returncode == 0  # not a fan of this
    assert res.data == ""
    # The regular module only logs "unable to fetch" with get_url


@pytest.mark.parametrize("template", (None, "jinja"))
@pytest.mark.parametrize("suffix", ("", "/"))
def test_get_dir(
    salt_ssh_cli_parameterized, tmp_path, template, suffix, master_cachedir
):
    tgt = tmp_path / ("many" if not template else "{{pillar.alot}}")
    res = salt_ssh_cli_parameterized.run(
        "cp.get_dir",
        "salt://" + ("grail" if not template else "{{pillar.script}}"),
        str(tgt) + suffix,
        template=template,
    )
    assert res.returncode == 0
    assert res.data
    assert isinstance(res.data, list)
    tgt = tmp_path / "many"
    master_path = master_cachedir / "files" / "base"
    for path in (tgt, master_path):
        assert path.exists()
        assert "grail" in os.listdir(path)
        assert "36" in os.listdir(path / "grail")
        assert "empty" in os.listdir(path / "grail")
        assert "scene" in os.listdir(path / "grail" / "36")
        if path == master_path:
            # otherwise we would include other cached files
            path = path / "grail"
            files = {str(master_path / Path(x).relative_to(tgt)) for x in res.data}
        else:
            files = set(res.data)
        # The regular cp.get_dir keeps superfluous path separators as well
        filelist = {
            str(x).replace(str(tgt), str(tgt) + suffix)
            for x in path.rglob("*")
            if not x.is_dir()
        }
        assert files == filelist


def test_get_dir_gzipped(salt_ssh_cli_parameterized, caplog, tmp_path):
    tgt = tmp_path / "many"
    res = salt_ssh_cli_parameterized.run("cp.get_dir", "salt://grail", tgt, gzip=5)
    assert "The gzip argument to cp.get_dir in salt-ssh is unsupported" in caplog.text
    assert res.returncode == 0
    assert res.data
    tgt = tmp_path / "many"
    assert isinstance(res.data, list)
    assert tgt.exists()
    assert "grail" in os.listdir(tgt)
    assert "36" in os.listdir(tgt / "grail")
    assert "empty" in os.listdir(tgt / "grail")
    assert "scene" in os.listdir(tgt / "grail" / "36")


def test_get_dir_nonexistent_source(salt_ssh_cli_parameterized, caplog):
    res = salt_ssh_cli_parameterized.run(
        "cp.get_dir", "salt://grail/non/ex/is/tent", ""
    )
    assert res.returncode == 0  # not a fan of this
    assert isinstance(res.data, list)
    assert not res.data


@pytest.mark.parametrize("dst_is_dir", (False, True))
def test_get_url(salt_ssh_cli_parameterized, tmp_path, dst_is_dir, master_cachedir):
    src = "salt://grail/scene33"
    if dst_is_dir:
        tgt = tmp_path
    else:
        tgt = tmp_path / "scene33"
    res = salt_ssh_cli_parameterized.run("cp.get_url", src, str(tgt))
    assert res.returncode == 0
    assert res.data
    tgt = tmp_path / "scene33"
    assert res.data == str(tgt)
    master_path = master_cachedir / "files" / "base" / "grail" / "scene33"
    for file in (tgt, master_path):
        assert file.exists()
        data = file.read_text(encoding="utf-8")
        assert "KNIGHT:  They're nervous, sire." in data
        assert "bacon" not in data


def test_get_url_makedirs(salt_ssh_cli_parameterized, tmp_path, master_cachedir):
    tgt = tmp_path / "make" / "dirs" / "scene33"
    res = salt_ssh_cli_parameterized.run(
        "cp.get_url", "salt://grail/scene33", str(tgt), makedirs=True
    )
    assert res.returncode == 0
    assert res.data
    assert res.data == str(tgt)
    master_path = master_cachedir / "files" / "base" / "grail" / "scene33"
    for file in (tgt, master_path):
        assert file.exists()
        data = file.read_text(encoding="utf-8")
        assert "KNIGHT:  They're nervous, sire." in data
        assert "bacon" not in data


def test_get_url_dest_empty(salt_ssh_cli_parameterized, cachedir, master_cachedir):
    """
    salt:// source and destination omitted, should still cache the file
    """
    res = salt_ssh_cli_parameterized.run("cp.get_url", "salt://grail/scene33")
    assert res.returncode == 0
    assert res.data
    assert isinstance(res.data, str)
    master_path = master_cachedir / "files" / "base" / "grail" / "scene33"
    tgt = _convert(cachedir, master_cachedir, master_path)
    assert res.data == str(tgt)
    for file in (tgt, master_path):
        assert file.exists()
        data = file.read_text(encoding="utf-8")
        assert "KNIGHT:  They're nervous, sire." in data
        assert "bacon" not in data


def test_get_url_no_dest(salt_ssh_cli_parameterized):
    """
    salt:// source given and destination set as None, should return the data
    """
    res = salt_ssh_cli_parameterized.run("cp.get_url", "salt://grail/scene33", None)
    assert res.returncode == 0
    assert res.data
    assert isinstance(res.data, str)
    assert "KNIGHT:  They're nervous, sire." in res.data
    assert "bacon" not in res.data


def test_get_url_nonexistent_source(salt_ssh_cli_parameterized, caplog):
    res = salt_ssh_cli_parameterized.run(
        "cp.get_url", "salt://grail/nonexistent_scene", None
    )
    assert res.returncode == 0  # not a fan of this
    assert res.data is False
    assert (
        "Unable to fetch file salt://grail/nonexistent_scene from saltenv base."
        in caplog.text
    )


def test_get_url_https(salt_ssh_cli_parameterized, tmp_path, master_cachedir):
    tgt = tmp_path / "index.html"
    res = salt_ssh_cli_parameterized.run(
        "cp.get_url", "https://saltproject.io/index.html", tgt
    )
    assert res.returncode == 0
    assert res.data
    assert res.data == str(tgt)
    master_path = (
        master_cachedir / "extrn_files" / "base" / "saltproject.io" / "index.html"
    )
    for path in (tgt, master_path):
        assert path.exists()
        data = path.read_text(encoding="utf-8")
        assert "Salt Project" in data


def test_get_url_https_dest_empty(
    salt_ssh_cli_parameterized, tmp_path, cachedir, master_cachedir
):
    """
    https:// source given and destination omitted, should still cache the file
    """
    res = salt_ssh_cli_parameterized.run(
        "cp.get_url", "https://saltproject.io/index.html"
    )
    assert res.returncode == 0
    assert res.data
    master_path = (
        master_cachedir / "extrn_files" / "base" / "saltproject.io" / "index.html"
    )
    tgt = _convert(cachedir, master_cachedir, master_path)
    assert res.data == str(tgt)
    for path in (tgt, master_path):
        assert path.exists()
        data = path.read_text(encoding="utf-8")
        assert "Salt Project" in data


def test_get_url_https_no_dest(salt_ssh_cli_parameterized):
    """
    https:// source given and destination set as None, should return the data
    """
    timeout = 500
    start = time.time()
    sleep = 5
    while time.time() - start <= timeout:
        res = salt_ssh_cli_parameterized.run(
            "cp.get_url", "https://saltproject.io/index.html", None
        )
        if isinstance(res.data, str) and res.data.find("HTTP 599") == -1:
            break
        time.sleep(sleep)
    if isinstance(res.data, str) and res.data.find("HTTP 599") != -1:
        raise Exception("https://saltproject.io/index.html returned 599 error")
    assert res.returncode == 0
    assert res.data
    assert isinstance(res.data, str)
    assert "Salt Project" in res.data


@pytest.mark.parametrize("scheme", ("file://", ""))
@pytest.mark.parametrize(
    "path,expected",
    (
        (Path(RUNTIME_VARS.FILES) / "file" / "base" / "file.big", True),
        (Path("_foo", "bar", "baz"), False),
    ),
)
def test_get_url_file(salt_ssh_cli_parameterized, path, expected, scheme):
    """
    Ensure the file:// scheme is not supported
    """
    res = salt_ssh_cli_parameterized.run("cp.get_url", scheme + str(path))
    assert res.returncode == 0
    assert res.data is False


def test_get_url_file_contents(salt_ssh_cli_parameterized, tmp_path, caplog):
    """
    A file:// source is not supported since it would need to fetch
    a file from the minion onto the master to be consistent
    """
    src = Path(RUNTIME_VARS.FILES) / "file" / "base" / "file.big"
    res = salt_ssh_cli_parameterized.run("cp.get_url", "file://" + str(src), None)
    assert res.returncode == 0
    assert res.data is False
    assert (
        "The file:// scheme is not supported via the salt-ssh cp wrapper" in caplog.text
    )


@pytest.mark.timeout(300)  # FTP can be slow, allow 5 minutes
def test_get_url_ftp(salt_ssh_cli_parameterized, tmp_path, master_cachedir):
    tgt = tmp_path / "README.TXT"
    res = salt_ssh_cli_parameterized.run(
        "cp.get_url", "ftp://ftp.freebsd.org/pub/FreeBSD/releases/amd64/README.TXT", tgt
    )
    assert res.returncode == 0
    assert res.data
    assert res.data == str(tgt)
    master_path = (
        master_cachedir
        / "extrn_files"
        / "base"
        / "ftp.freebsd.org"
        / "pub"
        / "FreeBSD"
        / "releases"
        / "amd64"
        / "README.TXT"
    )
    for path in (tgt, master_path):
        assert path.exists()
        data = path.read_text(encoding="utf-8")
        assert "The official FreeBSD" in data


def test_get_file_str_salt(salt_ssh_cli_parameterized, cachedir, master_cachedir):
    src = "salt://grail/scene33"
    res = salt_ssh_cli_parameterized.run("cp.get_file_str", src)
    assert res.returncode == 0
    assert res.data
    assert isinstance(res.data, str)
    assert "KNIGHT:  They're nervous, sire." in res.data
    tgt = cachedir / "files" / "base" / "grail" / "scene33"
    master_path = _convert(cachedir, master_cachedir, tgt, master=True)
    for path in (tgt, master_path):
        assert path.exists()
        text = path.read_text(encoding="utf-8")
        assert "KNIGHT:  They're nervous, sire." in text


def test_get_file_str_nonexistent_source(salt_ssh_cli_parameterized, caplog):
    src = "salt://grail/nonexistent_scene"
    res = salt_ssh_cli_parameterized.run("cp.get_file_str", src)
    assert res.returncode == 0  # yup...
    assert res.data is False


def test_get_file_str_https(salt_ssh_cli_parameterized, cachedir, master_cachedir):
    src = "https://saltproject.io/index.html"
    res = salt_ssh_cli_parameterized.run("cp.get_file_str", src)
    assert res.returncode == 0
    assert res.data
    assert isinstance(res.data, str)
    assert "Salt Project" in res.data
    tgt = cachedir / "extrn_files" / "base" / "saltproject.io" / "index.html"
    master_path = _convert(cachedir, master_cachedir, tgt, master=True)
    for path in (tgt, master_path):
        assert path.exists()
        text = path.read_text(encoding="utf-8")
        assert "Salt Project" in text


def test_get_file_str_local(salt_ssh_cli_parameterized, cachedir, caplog):
    src = Path(RUNTIME_VARS.FILES) / "file" / "base" / "cheese"
    res = salt_ssh_cli_parameterized.run("cp.get_file_str", "file://" + str(src))
    assert res.returncode == 0
    assert isinstance(res.data, str)
    assert "Gromit" in res.data
    assert (
        "The file:// scheme is not supported via the salt-ssh cp wrapper"
        not in caplog.text
    )


@pytest.mark.parametrize("suffix", ("", "?saltenv=prod"))
def test_cache_file(salt_ssh_cli_parameterized, suffix, cachedir, master_cachedir):
    res = salt_ssh_cli_parameterized.run("cp.cache_file", "salt://cheese" + suffix)
    assert res.returncode == 0
    assert res.data
    tgt = (
        cachedir
        / "files"
        / ("base" if "saltenv" not in suffix else suffix.split("=")[1])
        / "cheese"
    )
    master_path = _convert(cachedir, master_cachedir, tgt, master=True)
    for file in (tgt, master_path):
        data = file.read_text(encoding="utf-8")
        assert "Gromit" in data
        assert ("Comte" in data) is bool(suffix)


@pytest.fixture
def _cache_twice(
    salt_master, request, salt_ssh_cli_parameterized, cachedir, master_cachedir
):

    # ensure the cache is clean
    tgt = cachedir / "extrn_files" / "base" / "saltproject.io" / "index.html"
    tgt.unlink(missing_ok=True)
    master_tgt = _convert(cachedir, master_cachedir, tgt, master=True)
    master_tgt.unlink(missing_ok=True)

    # create a template that will cause a file to get cached twice
    # within the same context
    name = "cp_cache"
    src = "https://saltproject.io/index.html"
    remove = getattr(request, "param", False)
    contents = f"""
{{%- set cache = salt["cp.cache_file"]("{src}") %}}
{{%- if not cache %}}
{{#-   Stop rendering. It's one of the only ways to throw an exception
       during master-side rendering currently (in order to fail it).
#}}
{{%-   do salt["cp.get_file"]("foobar", template="thisthrowsanexception") %}}
{{%- endif %}}
{{%- set master_cache = salt["cp.convert_cache_path"](cache, master=true) %}}
{{%- do salt["file.append"](cache, "\nwasmodifiedhahaha") %}}
{{%- do salt["file.append"](master_cache, "\nwasmodifiedhahaha") %}}
"""
    if remove:
        contents += f"""
{{%- do salt["file.remove"]({'master_cache' if remove == 'master' else 'cache'}) %}}"""
    contents += f"""
{{%- set res2 = salt["cp.cache_file"]("{src}") %}}
{{{{ res2 }}}}
    """
    with salt_master.state_tree.base.temp_file(name, contents):
        yield f"salt://{name}"


def test_cache_file_context_cache(
    salt_ssh_cli_parameterized, cachedir, master_cachedir, _cache_twice
):
    res = salt_ssh_cli_parameterized.run(
        "slsutil.renderer", _cache_twice, default_renderer="jinja"
    )
    assert res.returncode == 0
    tgt = res.data.strip()
    assert tgt
    tgt = Path(tgt)
    for file in (tgt, _convert(cachedir, master_cachedir, tgt, master=True)):
        assert tgt.exists()
        # If both files were present, they should not be re-fetched
        assert "wasmodifiedhahaha" in tgt.read_text(encoding="utf-8")


@pytest.mark.parametrize("_cache_twice", ("master", "minion"), indirect=True)
def test_cache_file_context_cache_requires_both_caches(
    salt_ssh_cli_parameterized, cachedir, master_cachedir, _cache_twice
):
    res = salt_ssh_cli_parameterized.run(
        "slsutil.renderer", _cache_twice, default_renderer="jinja"
    )
    assert res.returncode == 0
    tgt = res.data.strip()
    assert tgt
    tgt = Path(tgt)
    for file in (tgt, _convert(cachedir, master_cachedir, tgt, master=True)):
        assert tgt.exists()
        # If one of the files was removed, it should be re-fetched
        assert "wasmodifiedhahaha" not in tgt.read_text(encoding="utf-8")


def test_cache_file_nonexistent_source(salt_ssh_cli_parameterized):
    res = salt_ssh_cli_parameterized.run(
        "cp.get_template", "salt://grail/nonexistent_scene", ""
    )
    assert res.returncode == 0  # not a fan of this
    assert res.data == ""
    # The regular module only logs "unable to fetch" with get_url


@pytest.mark.parametrize(
    "files",
    (
        ["salt://grail/scene33", "salt://grail/36/scene"],
        "salt://grail/scene33,salt://grail/36/scene",
    ),
)
def test_cache_files(salt_ssh_cli_parameterized, files):
    res = salt_ssh_cli_parameterized.run("cp.cache_files", files)
    assert res.returncode == 0
    assert res.data
    assert isinstance(res.data, list)
    for path in res.data:
        assert isinstance(path, str)
        path = Path(path)
        assert path.exists()
        data = Path(path).read_text(encoding="utf-8")
        assert "ARTHUR:" in data
        assert "bacon" not in data


def test_cache_dir(salt_ssh_cli_parameterized, cachedir, master_cachedir):
    res = salt_ssh_cli_parameterized.run("cp.cache_dir", "salt://grail")
    assert res.returncode == 0
    assert res.data
    assert isinstance(res.data, list)
    tgt = cachedir / "files" / "base" / "grail"
    master_path = _convert(cachedir, master_cachedir, tgt, master=True)
    for path in (tgt, master_path):
        assert path.exists()
        assert "36" in os.listdir(path)
        assert "empty" in os.listdir(path)
        assert "scene" in os.listdir(path / "36")
        if path == master_path:
            files = {str(master_path / Path(x).relative_to(tgt)) for x in res.data}
        else:
            files = set(res.data)
        filelist = {str(x) for x in path.rglob("*") if not x.is_dir()}
        assert files == filelist


def test_cache_dir_nonexistent_source(salt_ssh_cli_parameterized, caplog):
    res = salt_ssh_cli_parameterized.run(
        "cp.cache_dir", "salt://grail/non/ex/is/tent", ""
    )
    assert res.returncode == 0  # not a fan of this
    assert isinstance(res.data, list)
    assert not res.data


def test_list_states(salt_master, salt_ssh_cli_parameterized, tmp_path):
    top_sls = """
    base:
      '*':
        - core
        """
    core_state = f"""
    {tmp_path / "testfile"}/testfile:
      file.managed:
        - source: salt://testfile
        - makedirs: true
        """

    with salt_master.state_tree.base.temp_file(
        "top.sls", top_sls
    ), salt_master.state_tree.base.temp_file("core.sls", core_state):
        res = salt_ssh_cli_parameterized.run(
            "cp.list_states",
        )
        assert res.returncode == 0
        assert res.data
        assert isinstance(res.data, list)
        assert "core" in res.data
        assert "top" in res.data
        assert "cheese" not in res.data


def test_list_master(salt_ssh_cli_parameterized):
    res = salt_ssh_cli_parameterized.run("cp.list_master")
    assert res.returncode == 0
    assert res.data
    assert isinstance(res.data, list)
    for file in [
        "cheese",
        "grail/empty",
        "grail/36/scene",
        "_modules/salttest.py",
        "running.sls",
        "test_deep/a/test.sls",
    ]:
        assert file in res.data
    assert "test_deep/a" not in res.data


def test_list_master_dirs(salt_ssh_cli_parameterized):
    res = salt_ssh_cli_parameterized.run("cp.list_master_dirs")
    assert res.returncode == 0
    assert res.data
    assert isinstance(res.data, list)
    for path in ["test_deep", "test_deep/a", "test_deep/b/2"]:
        assert path in res.data
    for path in [
        "test_deep/test.sls",
        "test_deep/a/test.sls",
        "test_deep/b/2/test.sls",
        "cheese",
    ]:
        assert path not in res.data


def test_list_master_symlinks(salt_ssh_cli_parameterized, salt_master):
    if salt_ssh_cli_parameterized.config.get("fileserver_ignoresymlinks", False):
        pytest.skip("Fileserver is configured to ignore symlinks")
    with salt_master.state_tree.base.temp_file(random_string("foo-"), "") as tgt:
        sym = tgt.parent / random_string("test_list_master_symlinks-")
        try:
            sym.symlink_to(tgt)
            res = salt_ssh_cli_parameterized.run("cp.list_master_symlinks")
            assert res.returncode == 0
            assert res.data
            assert isinstance(res.data, dict)
            assert res.data
            assert sym.name in res.data
            assert res.data[sym.name] == str(tgt)
        finally:
            # Clean up the symlink to avoid conflicts with parametrized tests
            if sym.exists() or sym.is_symlink():
                sym.unlink()


@pytest.fixture(params=(False, "cached", "render_cached"))
def _is_cached(salt_ssh_cli_parameterized, suffix, request, cachedir, master_cachedir):
    remove = ["files", "extrn_files"]
    if request.param == "cached":
        ret = salt_ssh_cli_parameterized.run(
            "cp.cache_file", "salt://grail/scene33" + suffix
        )
        assert ret.returncode == 0
        assert ret.data
        remove.remove("files")
    elif request.param == "render_cached":
        ret = salt_ssh_cli_parameterized.run(
            "cp.get_template", "salt://grail/scene33" + suffix, "", spam="bacon"
        )
        assert ret.returncode == 0
        assert ret.data
        remove.remove("extrn_files")
    for basedir in remove:
        tgt = cachedir / basedir / "base" / "grail" / "scene33"
        tgt.unlink(missing_ok=True)
        master_tgt = _convert(cachedir, master_cachedir, tgt, master=True)
        master_tgt.unlink(missing_ok=True)
    return request.param


@pytest.mark.parametrize("suffix", ("", "?saltenv=base"))
def test_is_cached(salt_ssh_cli_parameterized, cachedir, _is_cached, suffix):
    """
    is_cached should find both cached files from the fileserver as well
    as cached rendered templates
    """
    if _is_cached == "render_cached":
        tgt = cachedir / "extrn_files" / "base" / "grail" / "scene33"
    else:
        tgt = cachedir / "files" / "base" / "grail" / "scene33"
    res = salt_ssh_cli_parameterized.run(
        "cp.is_cached", "salt://grail/scene33" + suffix
    )
    assert res.returncode == 0
    assert (res.data == str(tgt)) is bool(_is_cached)
    assert (res.data != "") is bool(_is_cached)


def test_is_cached_nonexistent(salt_ssh_cli_parameterized):
    res2 = salt_ssh_cli_parameterized.run("cp.is_cached", "salt://fasldkgj/poicxzbn")
    assert res2.returncode == 0
    assert res2.data == ""


@pytest.mark.parametrize("suffix", ("", "?saltenv=base"))
def test_hash_file(salt_ssh_cli_parameterized, cachedir, master_cachedir, suffix):
    res = salt_ssh_cli_parameterized.run(
        "cp.hash_file", "salt://grail/scene33" + suffix
    )
    assert res.returncode == 0
    assert res.data
    sha256_hash = res.data["hsum"]
    res = salt_ssh_cli_parameterized.run("cp.cache_file", "salt://grail/scene33")
    assert res.returncode == 0
    assert res.data
    master_path = _convert(cachedir, master_cachedir, res.data, master=True)
    assert master_path.exists()
    data = master_path.read_bytes()
    digest = hashlib.sha256(data).hexdigest()
    assert digest == sha256_hash


def test_hash_file_local(salt_ssh_cli_parameterized, caplog):
    """
    Ensure that local files are run through ``salt-call`` on the target.
    We have to trust that this would otherwise fail because the tests
    run against localhost.
    """
    path = Path(RUNTIME_VARS.FILES) / "file" / "base" / "cheese"
    res = salt_ssh_cli_parameterized.run("cp.hash_file", str(path))
    assert res.returncode == 0
    # This would be logged if SSHCpClient was used instead of
    # performing a shimmed salt-call command
    assert "Hashing local files is not supported via salt-ssh" not in caplog.text
    assert isinstance(res.data, dict)
    assert res.data
    data = path.read_bytes()
    digest = hashlib.sha256(data).hexdigest()
    sha256_hash = res.data["hsum"]
    assert digest == sha256_hash


@pytest.fixture
def state_tree_jinjaimport(tmp_path, salt_master):
    map_contents = """{%- set mapdata = {"foo": "bar"} %}"""
    managed_contents = """
        {%- from "my/map.jinja" import mapdata with context %}
        {{- mapdata["foo"] -}}
    """
    state_contents = f"""
{{%- do salt["cp.cache_file"]("salt://my/map.jinja") %}}

Serialize config:
  file.managed:
    - name: {tmp_path / "config.conf"}
    - source: salt://my/files/config.conf.j2
    - template: jinja
"""
    with salt_master.state_tree.base.temp_file(
        "my/file_managed_import.sls", state_contents
    ) as state, salt_master.state_tree.base.temp_file(
        "my/map.jinja", map_contents
    ), salt_master.state_tree.base.temp_file(
        "my/files/config.conf.j2", managed_contents
    ):
        yield f"my.{state.stem}"


def test_cp_cache_file_as_workaround_for_missing_map_file(
    salt_ssh_cli_parameterized, state_tree_jinjaimport, tmp_path
):
    tgt = tmp_path / "config.conf"
    ret = salt_ssh_cli_parameterized.run("state.sls", state_tree_jinjaimport)
    assert ret.returncode == 0
    assert isinstance(ret.data, dict)
    assert ret.data
    assert tgt.exists()
    assert tgt.read_text(encoding="utf-8").strip() == "bar"



import pytest

from tests.support.helpers import system_python_version

pytestmark = [
    pytest.mark.slow_test,
    pytest.mark.skipif(
        system_python_version() < (3, 10),
        reason="System python too old for these tests",
    ),
]


@pytest.fixture(scope="module", autouse=True)
def pillar_tree(base_env_pillar_tree_root_dir):
    top_file = """
    base:
      'localhost':
        - basic
      '127.0.0.1':
        - basic
    """
    basic_pillar_file = """
    alot: many
    """
    top_tempfile = pytest.helpers.temp_file(
        "top.sls", top_file, base_env_pillar_tree_root_dir
    )
    basic_tempfile = pytest.helpers.temp_file(
        "basic.sls", basic_pillar_file, base_env_pillar_tree_root_dir
    )

    with top_tempfile, basic_tempfile:
        yield


@pytest.fixture(scope="module")
def script_templated(base_env_state_tree_root_dir):
    contents = """
#!/usr/bin/env bash

echo {{ pillar["alot"] }}
"""
    with pytest.helpers.temp_file(
        "parrot.sh", contents, base_env_state_tree_root_dir
    ) as script:
        yield f"salt://{script.name}"


def test_script(salt_ssh_cli):
    args = "saltines crackers biscuits=yes"
    script = "salt://script.py"
    ret = salt_ssh_cli.run("cmd.script", script, args)
    assert ret.returncode == 0
    assert isinstance(ret.data, dict)
    assert ret.data
    assert ret.data["stdout"] == args


def test_script_query_string(salt_ssh_cli):
    args = "saltines crackers biscuits=yes"
    script = "salt://script.py?saltenv=base"
    ret = salt_ssh_cli.run("cmd.script", script, args)
    assert ret.returncode == 0
    assert isinstance(ret.data, dict)
    assert ret.data
    assert ret.data["stdout"] == args


def test_script_cwd(salt_ssh_cli, tmp_path):
    args = "saltines crackers biscuits=yes"
    script = "salt://script.py"
    # can't pass cwd as kwarg
    ret = salt_ssh_cli.run("cmd.script", script, args, tmp_path)
    assert ret.returncode == 0
    assert isinstance(ret.data, dict)
    assert ret.data
    assert ret.data["stdout"] == args


def test_script_cwd_with_space(salt_ssh_cli, tmp_path):
    tmp_cwd = tmp_path / "test 2"
    tmp_cwd.mkdir()
    args = "saltines crackers biscuits=yes"
    script = "salt://script.py"
    ret = salt_ssh_cli.run("cmd.script", script, args, cwd=tmp_cwd)
    assert ret.returncode == 0
    assert isinstance(ret.data, dict)
    assert ret.data
    assert ret.data["stdout"] == args


@pytest.mark.parametrize("template", (None, "jinja"))
def test_script_nonexistent(salt_ssh_cli, template):
    script = "salt://non/ex/is/tent.sh"
    ret = salt_ssh_cli.run("cmd.script", script, "", template=template)
    assert ret.returncode == 0  # meh
    assert isinstance(ret.data, dict)
    assert ret.data
    assert "cache_error" in ret.data
    assert "retcode" in ret.data
    assert ret.data["retcode"] == 1


@pytest.mark.parametrize("pillar", (None, {"alot": "meow"}))
def test_script_template(salt_ssh_cli, script_templated, pillar):
    ret = salt_ssh_cli.run(
        "cmd.script", script_templated, template="jinja", pillar=pillar
    )
    assert ret.returncode == 0
    assert isinstance(ret.data, dict)
    assert ret.data
    assert ret.data["stdout"] == (pillar or {}).get("alot", "many")


def test_script_retcode(salt_ssh_cli):
    script = "salt://script.py"
    ret = salt_ssh_cli.run("cmd.script_retcode", script)
    assert ret.returncode == 0
    assert ret.data == 0



"""
Verify salt-ssh does not report success when it cannot parse
the return value.
"""

import pytest

from salt.defaults.exitcodes import EX_AGGREGATE
from tests.support.helpers import system_python_version

pytestmark = [
    pytest.mark.skip_on_windows(reason="salt-ssh not available on Windows"),
    pytest.mark.slow_test,
    pytest.mark.skipif(
        system_python_version() < (3, 10),
        reason="System python too old for these tests",
    ),
]


@pytest.fixture(scope="module")
def state_tree_remote_exception_mod(salt_run_cli, base_env_state_tree_root_dir):
    module_contents = r"""
import os
import sys


def __virtual__():
    return "whoops"


def do_stuff(name):
    print("Hi there, nice that you're trying to make me do stuff.", file=sys.stderr)
    print("Traceback (most recent call last):", file=sys.stderr)
    print('  File "/dont/treat/me/like/that.py" in buzz_off', file=sys.stderr)
    print("ComplianceError: 'Outlaw detected'", file=sys.stderr)
    sys.stderr.flush()
    os._exit(1)
"""
    module_dir = base_env_state_tree_root_dir / "_states"
    module_tempfile = pytest.helpers.temp_file("whoops.py", module_contents, module_dir)
    try:
        with module_tempfile:
            ret = salt_run_cli.run("saltutil.sync_states")
            assert ret.returncode == 0
            assert "states.whoops" in ret.data
            yield
    finally:
        ret = salt_run_cli.run("saltutil.sync_states")
        assert ret.returncode == 0


@pytest.fixture(scope="module", autouse=True)
def state_tree_remote_exception(
    base_env_state_tree_root_dir, state_tree_remote_exception_mod
):
    top_file = """
    base:
      'localhost':
        - remote_stacktrace
      '127.0.0.1':
        - remote_stacktrace
    """
    state_file = r"""
    This should be detected as failure:
      whoops.do_stuff
    """
    top_tempfile = pytest.helpers.temp_file(
        "top.sls", top_file, base_env_state_tree_root_dir
    )
    state_tempfile = pytest.helpers.temp_file(
        "remote_stacktrace.sls", state_file, base_env_state_tree_root_dir
    )
    with top_tempfile, state_tempfile:
        yield


@pytest.mark.slow_test
@pytest.mark.usefixtures("state_tree_remote_exception")
@pytest.mark.parametrize(
    "args",
    (
        ("state.sls", "remote_stacktrace"),
        ("state.highstate",),
        ("state.sls_id", "This should be detected as failure", "remote_stacktrace"),
        ("state.top", "top.sls"),
        ("state.single", "whoops.do_stuff", "now"),
    ),
)
def test_it(salt_ssh_cli_parameterized, ssh_deployment_type, args):
    ret = salt_ssh_cli_parameterized.run(*args)

    assert ret.returncode == EX_AGGREGATE
    assert ret.data
    assert isinstance(ret.data, dict)
    assert "ComplianceError: 'Outlaw detected'" in ret.data["stderr"]



import pytest

import salt.modules.macdefaults as macdefaults_module
import salt.states.macdefaults as macdefaults
from tests.support.mock import MagicMock, patch


@pytest.fixture
def configure_loader_modules():
    return {
        macdefaults: {
            "__salt__": {
                "macdefaults.cast_value_to_vtype": macdefaults_module.cast_value_to_vtype
            }
        }
    }


def test_write_default():
    """
    Test writing a default setting
    """
    expected = {
        "changes": {"written": "com.apple.CrashReporter DialogType is set to Server"},
        "comment": "",
        "name": "DialogType",
        "result": True,
    }

    read_mock = MagicMock(return_value="Local")
    write_mock = MagicMock(return_value={"retcode": 0})
    with patch.dict(
        macdefaults.__salt__,
        {"macdefaults.read": read_mock, "macdefaults.write": write_mock},
    ):
        out = macdefaults.write("DialogType", "com.apple.CrashReporter", "Server")
        read_mock.assert_called_once_with(
            "com.apple.CrashReporter", "DialogType", None, key_separator=None
        )
        write_mock.assert_called_once_with(
            "com.apple.CrashReporter",
            "DialogType",
            "Server",
            None,
            None,
            key_separator=None,
        )
        assert out == expected


def test_write_default_already_set():
    """
    Test writing a default setting that is already set
    """
    key = "DialogType.Value"

    expected = {
        "changes": {},
        "comment": f"com.apple.CrashReporter {key} is already set to Server",
        "name": key,
        "result": True,
    }

    read_mock = MagicMock(return_value="Server")
    write_mock = MagicMock(return_value={"retcode": 0})
    with patch.dict(
        macdefaults.__salt__,
        {"macdefaults.read": read_mock, "macdefaults.write": write_mock},
    ):
        out = macdefaults.write(
            key, "com.apple.CrashReporter", "Server", name_separator="."
        )
        read_mock.assert_called_once_with(
            "com.apple.CrashReporter", key, None, key_separator="."
        )
        assert not write_mock.called
        assert out == expected


def test_write_default_boolean():
    """
    Test writing a default setting with a boolean
    """
    expected = {
        "changes": {"written": "com.apple.something Key is set to True"},
        "comment": "",
        "name": "Key",
        "result": True,
    }

    read_mock = MagicMock(return_value=False)
    write_mock = MagicMock(return_value={"retcode": 0})
    with patch.dict(
        macdefaults.__salt__,
        {"macdefaults.read": read_mock, "macdefaults.write": write_mock},
    ):
        out = macdefaults.write("Key", "com.apple.something", "YES", vtype="boolean")
        read_mock.assert_called_once_with(
            "com.apple.something", "Key", None, key_separator=None
        )
        write_mock.assert_called_once_with(
            "com.apple.something", "Key", True, "boolean", None, key_separator=None
        )
        assert out == expected


def test_write_default_boolean_already_set():
    """
    Test writing a default setting with a boolean that is already set
    """
    expected = {
        "changes": {},
        "comment": "com.apple.something Key is already set to True",
        "name": "Key",
        "result": True,
    }

    read_mock = MagicMock(return_value=True)
    write_mock = MagicMock(return_value={"retcode": 0})
    with patch.dict(
        macdefaults.__salt__,
        {"macdefaults.read": read_mock, "macdefaults.write": write_mock},
    ):
        out = macdefaults.write("Key", "com.apple.something", "YES", vtype="boolean")
        read_mock.assert_called_once_with(
            "com.apple.something", "Key", None, key_separator=None
        )
        assert not write_mock.called
        assert out == expected


def test_write_default_integer():
    """
    Test writing a default setting with a integer
    """
    expected = {
        "changes": {"written": "com.apple.something Key is set to 1337"},
        "comment": "",
        "name": "Key",
        "result": True,
    }

    read_mock = MagicMock(return_value=99)
    write_mock = MagicMock(return_value={"retcode": 0})
    with patch.dict(
        macdefaults.__salt__,
        {"macdefaults.read": read_mock, "macdefaults.write": write_mock},
    ):
        out = macdefaults.write("Key", "com.apple.something", 1337)
        read_mock.assert_called_once_with(
            "com.apple.something", "Key", None, key_separator=None
        )
        write_mock.assert_called_once_with(
            "com.apple.something", "Key", 1337, None, None, key_separator=None
        )
        assert out == expected


def test_write_default_integer_already_set():
    """
    Test writing a default setting with an integer that is already set
    """
    key = "Key.subKey"

    expected = {
        "changes": {},
        "comment": f"com.apple.something {key} is already set to 1337",
        "name": key,
        "result": True,
    }

    read_mock = MagicMock(return_value=1337)
    write_mock = MagicMock(return_value={"retcode": 0})
    with patch.dict(
        macdefaults.__salt__,
        {"macdefaults.read": read_mock, "macdefaults.write": write_mock},
    ):
        out = macdefaults.write(
            key,
            "com.apple.something",
            1337,
            vtype="integer",
            name_separator=".",
        )
        read_mock.assert_called_once_with(
            "com.apple.something", key, None, key_separator="."
        )
        assert not write_mock.called
        assert out == expected


def test_write_default_float():
    """
    Test writing a default setting with a float
    """
    expected = {
        "changes": {"written": "com.apple.something Key is set to 0.865"},
        "comment": "",
        "name": "Key",
        "result": True,
    }

    read_mock = MagicMock(return_value=0.4)
    write_mock = MagicMock(return_value={"retcode": 0})
    with patch.dict(
        macdefaults.__salt__,
        {"macdefaults.read": read_mock, "macdefaults.write": write_mock},
    ):
        out = macdefaults.write("Key", "com.apple.something", "0.8650", vtype="float")
        read_mock.assert_called_once_with(
            "com.apple.something", "Key", None, key_separator=None
        )
        write_mock.assert_called_once_with(
            "com.apple.something", "Key", 0.865, "float", None, key_separator=None
        )
        assert out == expected


def test_write_default_float_already_set():
    """
    Test writing a default setting with a float that is already set_default
    """
    expected = {
        "changes": {},
        "comment": "com.apple.something Key is already set to 0.865",
        "name": "Key",
        "result": True,
    }

    read_mock = MagicMock(return_value=0.865)
    write_mock = MagicMock(return_value={"retcode": 0})
    with patch.dict(
        macdefaults.__salt__,
        {"macdefaults.read": read_mock, "macdefaults.write": write_mock},
    ):
        out = macdefaults.write("Key", "com.apple.something", 0.86500)
        read_mock.assert_called_once_with(
            "com.apple.something", "Key", None, key_separator=None
        )
        assert not write_mock.called
        assert out == expected


def test_write_default_array():
    """
    Test writing a default setting with an array
    """
    key = "Key.subKey"

    value = ["a", 1, 0.5, True]
    expected = {
        "changes": {"written": f"com.apple.something {key} is set to {value}"},
        "comment": "",
        "name": key,
        "result": True,
    }

    read_mock = MagicMock(return_value=None)
    write_mock = MagicMock(return_value={"retcode": 0})
    with patch.dict(
        macdefaults.__salt__,
        {"macdefaults.read": read_mock, "macdefaults.write": write_mock},
    ):
        out = macdefaults.write(
            key,
            "com.apple.something",
            value,
            vtype="array",
            name_separator=".",
        )
        read_mock.assert_called_once_with(
            "com.apple.something", key, None, key_separator="."
        )
        write_mock.assert_called_once_with(
            "com.apple.something", key, value, "array", None, key_separator="."
        )
        assert out == expected


def test_write_default_array_already_set():
    """
    Test writing a default setting with an array that is already set
    """
    value = ["a", 1, 0.5, True]
    expected = {
        "changes": {},
        "comment": f"com.apple.something Key is already set to {value}",
        "name": "Key",
        "result": True,
    }

    read_mock = MagicMock(return_value=value)
    write_mock = MagicMock(return_value={"retcode": 0})
    with patch.dict(
        macdefaults.__salt__,
        {"macdefaults.read": read_mock, "macdefaults.write": write_mock},
    ):
        out = macdefaults.write("Key", "com.apple.something", value)
        read_mock.assert_called_once_with(
            "com.apple.something", "Key", None, key_separator=None
        )
        assert not write_mock.called
        assert out == expected


def test_write_default_array_add():
    """
    Test writing a default setting adding an array to another
    """
    write_value = ["a", 1]
    read_value = ["b", 2]
    expected = {
        "changes": {"written": f"com.apple.something Key is set to {write_value}"},
        "comment": "",
        "name": "Key",
        "result": True,
    }

    read_mock = MagicMock(return_value=read_value)
    write_mock = MagicMock(return_value={"retcode": 0})
    with patch.dict(
        macdefaults.__salt__,
        {"macdefaults.read": read_mock, "macdefaults.write": write_mock},
    ):
        out = macdefaults.write(
            "Key", "com.apple.something", write_value, vtype="array-add"
        )
        read_mock.assert_called_once_with(
            "com.apple.something", "Key", None, key_separator=None
        )
        write_mock.assert_called_once_with(
            "com.apple.something",
            "Key",
            write_value,
            "array-add",
            None,
            key_separator=None,
        )
        assert out == expected


def test_write_default_array_add_already_set_distinct_order():
    """
    Test writing a default setting adding an array to another that is already set
    The new array is in a different order than the last elements of the existing one
    """
    write_value = [2, "a"]
    read_value = ["b", 1, "a", 2]
    expected = {
        "changes": {"written": f"com.apple.something Key is set to {write_value}"},
        "comment": "",
        "name": "Key",
        "result": True,
    }

    read_mock = MagicMock(return_value=read_value)
    write_mock = MagicMock(return_value={"retcode": 0})
    with patch.dict(
        macdefaults.__salt__,
        {"macdefaults.read": read_mock, "macdefaults.write": write_mock},
    ):
        out = macdefaults.write(
            "Key", "com.apple.something", write_value, vtype="array-add"
        )
        read_mock.assert_called_once_with(
            "com.apple.something", "Key", None, key_separator=None
        )
        write_mock.assert_called_once_with(
            "com.apple.something",
            "Key",
            write_value,
            "array-add",
            None,
            key_separator=None,
        )
        assert out == expected


def test_write_default_array_add_already_set_same_order():
    """
    Test writing a default setting adding an array to another that is already set
    The new array is already in the same order as the existing one
    """
    write_value = [1, 2]
    read_value = ["b", "a", 1, 2]
    expected = {
        "changes": {},
        "comment": f"com.apple.something Key is already set to {write_value}",
        "name": "Key",
        "result": True,
    }

    read_mock = MagicMock(return_value=read_value)
    write_mock = MagicMock(return_value={"retcode": 0})
    with patch.dict(
        macdefaults.__salt__,
        {"macdefaults.read": read_mock, "macdefaults.write": write_mock},
    ):
        out = macdefaults.write(
            "Key", "com.apple.something", write_value, vtype="array-add"
        )
        read_mock.assert_called_once_with(
            "com.apple.something", "Key", None, key_separator=None
        )
        assert not write_mock.called
        assert out == expected


def test_write_default_dict():
    """
    Test writing a default setting with a dictionary
    """
    value = {"string": "bar", "integer": 1, "float": 0.5, "boolean": True}
    expected = {
        "changes": {"written": f"com.apple.something Key is set to {value}"},
        "comment": "",
        "name": "Key",
        "result": True,
    }

    read_mock = MagicMock(return_value=None)
    write_mock = MagicMock(return_value={"retcode": 0})
    with patch.dict(
        macdefaults.__salt__,
        {"macdefaults.read": read_mock, "macdefaults.write": write_mock},
    ):
        out = macdefaults.write("Key", "com.apple.something", value, vtype="dict")
        read_mock.assert_called_once_with(
            "com.apple.something", "Key", None, key_separator=None
        )
        write_mock.assert_called_once_with(
            "com.apple.something", "Key", value, "dict", None, key_separator=None
        )
        assert out == expected


def test_write_default_dict_already_set():
    """
    Test writing a default setting with a dictionary that is already set
    """
    value = {"string": "bar", "integer": 1, "float": 0.5, "boolean": True}
    expected = {
        "changes": {},
        "comment": f"com.apple.something Key is already set to {value}",
        "name": "Key",
        "result": True,
    }

    read_mock = MagicMock(return_value=value)
    write_mock = MagicMock(return_value={"retcode": 0})
    with patch.dict(
        macdefaults.__salt__,
        {"macdefaults.read": read_mock, "macdefaults.write": write_mock},
    ):
        out = macdefaults.write("Key", "com.apple.something", value, vtype="dict")
        read_mock.assert_called_once_with(
            "com.apple.something", "Key", None, key_separator=None
        )
        assert not write_mock.called
        assert out == expected


def test_write_default_dict_add():
    """
    Test writing a default setting adding elements to a dictionary
    """
    write_value = {"string": "bar", "integer": 1}
    read_value = {"integer": 1, "float": 0.5, "boolean": True}
    expected = {
        "changes": {"written": f"com.apple.something Key is set to {write_value}"},
        "comment": "",
        "name": "Key",
        "result": True,
    }

    read_mock = MagicMock(return_value=read_value)
    write_mock = MagicMock(return_value={"retcode": 0})
    with patch.dict(
        macdefaults.__salt__,
        {"macdefaults.read": read_mock, "macdefaults.write": write_mock},
    ):
        out = macdefaults.write(
            "Key", "com.apple.something", write_value, vtype="dict-add"
        )
        read_mock.assert_called_once_with(
            "com.apple.something", "Key", None, key_separator=None
        )
        write_mock.assert_called_once_with(
            "com.apple.something",
            "Key",
            write_value,
            "dict-add",
            None,
            key_separator=None,
        )
        assert out == expected


def test_write_default_dict_add_already_set():
    """
    Test writing a default setting adding elements to a dictionary that is already set
    """
    write_value = {"string": "bar", "integer": 1}
    read_value = {"string": "bar", "integer": 1, "float": 0.5, "boolean": True}
    expected = {
        "changes": {},
        "comment": f"com.apple.something Key is already set to {write_value}",
        "name": "Key",
        "result": True,
    }

    read_mock = MagicMock(return_value=read_value)
    write_mock = MagicMock(return_value={"retcode": 0})
    with patch.dict(
        macdefaults.__salt__,
        {"macdefaults.read": read_mock, "macdefaults.write": write_mock},
    ):
        out = macdefaults.write(
            "Key", "com.apple.something", write_value, vtype="dict-add"
        )
        read_mock.assert_called_once_with(
            "com.apple.something", "Key", None, key_separator=None
        )
        assert not write_mock.called
        assert out == expected


def test_absent_default_already():
    """
    Test ensuring non-existent defaults value is absent
    """
    expected = {
        "changes": {},
        "comment": "com.apple.something Key is already absent",
        "name": "Key",
        "result": True,
    }

    mock = MagicMock(return_value=None)
    with patch.dict(macdefaults.__salt__, {"macdefaults.delete": mock}):
        out = macdefaults.absent("Key", "com.apple.something")
        mock.assert_called_once_with(
            "com.apple.something", "Key", None, key_separator=None
        )
        assert out == expected


def test_absent_default_deleting_existing():
    """
    Test removing an existing default value
    """
    key = "Key.subKey"

    expected = {
        "changes": {"absent": f"com.apple.something {key} is now absent"},
        "comment": "",
        "name": key,
        "result": True,
    }

    mock = MagicMock(return_value={"retcode": 0})
    with patch.dict(macdefaults.__salt__, {"macdefaults.delete": mock}):
        out = macdefaults.absent(key, "com.apple.something", name_separator=".")
        mock.assert_called_once_with(
            "com.apple.something", key, None, key_separator="."
        )
        assert out == expected



import logging
import os

import pytest

import salt.serializers.json as jsonserializer
import salt.serializers.msgpack as msgpackserializer
import salt.serializers.yaml as yamlserializer
import salt.states.file as filestate
import salt.utils.files
import salt.utils.json
import salt.utils.platform
import salt.utils.win_functions
import salt.utils.yaml
from tests.support.mock import MagicMock, mock_open, patch

log = logging.getLogger(__name__)


@pytest.fixture
def configure_loader_modules():
    return {
        filestate: {
            "__env__": "base",
            "__salt__": {"file.manage_file": False},
            "__serializers__": {
                "yaml.serialize": yamlserializer.serialize,
                "yaml.seserialize": yamlserializer.serialize,
                "json.serialize": jsonserializer.serialize,
                "msgpack.serialize": msgpackserializer.serialize,
            },
            "__opts__": {"test": False, "cachedir": ""},
            "__instance_id__": "",
            "__low__": {},
            "__utils__": {},
        }
    }


# 'comment' function tests: 1
def test_comment():
    """
    Test to comment out specified lines in a file.
    """
    with patch.object(os.path, "exists", MagicMock(return_value=True)):
        name = "/etc/aliases" if salt.utils.platform.is_darwin() else "/etc/fstab"
        regex = "bind 127.0.0.1"

        ret = {"name": name, "result": False, "comment": "", "changes": {}}

        comt = "Must provide name to file.comment"
        ret.update({"comment": comt, "name": ""})
        assert filestate.comment("", regex) == ret

        mock_t = MagicMock(return_value=True)
        mock_f = MagicMock(return_value=False)
        with patch.object(os.path, "isabs", mock_f):
            comt = f"Specified file {name} is not an absolute path"
            ret.update({"comment": comt, "name": name})
            assert filestate.comment(name, regex) == ret

        with patch.object(os.path, "isabs", mock_t):
            with patch.dict(
                filestate.__salt__,
                {
                    "file.search": MagicMock(
                        side_effect=[False, True, False, False, False, False]
                    )
                },
            ):
                comt = "Pattern already commented"
                ret.update({"comment": comt, "result": True})
                assert filestate.comment(name, regex) == ret

                comt = "Pattern not found and ignore_missing set to True"
                ret.update({"comment": comt, "result": True})
                assert filestate.comment(name, regex, ignore_missing=True) == ret

                comt = f"{regex}: Pattern not found"
                ret.update({"comment": comt, "result": False})
                assert filestate.comment(name, regex) == ret

            with patch.dict(
                filestate.__salt__,
                {
                    "file.search": MagicMock(
                        side_effect=[True, True, True, False, True]
                    ),
                    "file.comment": mock_t,
                    "file.comment_line": mock_t,
                },
            ):
                with patch.dict(filestate.__opts__, {"test": True}):
                    comt = f"File {name} is set to be updated"
                    ret.update(
                        {"comment": comt, "result": None, "changes": {name: "updated"}}
                    )
                    assert filestate.comment(name, regex) == ret

                with patch.dict(filestate.__opts__, {"test": False}):
                    with patch.object(
                        salt.utils.files, "fopen", MagicMock(mock_open())
                    ):
                        comt = "Commented lines successfully"
                        ret.update({"comment": comt, "result": True, "changes": {}})
                        assert filestate.comment(name, regex) == ret

                with patch.dict(filestate.__opts__, {"test": True}):
                    comt = "Pattern already commented"
                    ret.update({"comment": comt, "result": True, "changes": {}})
                    assert filestate.comment(name, regex) == ret


# 'uncomment' function tests: 1
def test_uncomment():
    """
    Test to uncomment specified commented lines in a file
    """
    with patch.object(os.path, "exists", MagicMock(return_value=True)):
        name = "/etc/aliases" if salt.utils.platform.is_darwin() else "/etc/fstab"
        regex = "bind 127.0.0.1"

        ret = {"name": name, "result": False, "comment": "", "changes": {}}

        comt = "Must provide name to file.uncomment"
        ret.update({"comment": comt, "name": ""})
        assert filestate.uncomment("", regex) == ret

        mock_t = MagicMock(return_value=True)
        mock_f = MagicMock(return_value=False)
        mock = MagicMock(side_effect=[False, True, False, False, True, True, True])
        with patch.object(os.path, "isabs", mock_f):
            comt = f"Specified file {name} is not an absolute path"
            ret.update({"comment": comt, "name": name})
            assert filestate.uncomment(name, regex) == ret

        with patch.object(os.path, "isabs", mock_t):
            with patch.dict(
                filestate.__salt__,
                {
                    "file.search": mock,
                    "file.uncomment": mock_t,
                    "file.comment_line": mock_t,
                },
            ):
                comt = "Pattern already uncommented"
                ret.update({"comment": comt, "result": True})
                assert filestate.uncomment(name, regex) == ret

                comt = f"{regex}: Pattern not found"
                ret.update({"comment": comt, "result": False})
                assert filestate.uncomment(name, regex) == ret

                with patch.dict(filestate.__opts__, {"test": True}):
                    comt = f"File {name} is set to be updated"
                    ret.update(
                        {"comment": comt, "result": None, "changes": {name: "updated"}}
                    )
                    assert filestate.uncomment(name, regex) == ret

                with patch.dict(filestate.__opts__, {"test": False}):
                    with patch.object(
                        salt.utils.files, "fopen", MagicMock(mock_open())
                    ):
                        comt = "Uncommented lines successfully"
                        ret.update({"comment": comt, "result": True, "changes": {}})
                        assert filestate.uncomment(name, regex) == ret



"""
Manage OpenSSH keys and certificates
====================================

.. versionadded:: 3008.0


:depends: cryptography

.. note::

    * Certificate operations require at least cryptography release 40
    * Certificate operations with ``force-command`` and ``source-address``
      (or any critical option/extension with a value) should only be relied
      upon in releases >= 41.0.2 (https://github.com/pyca/cryptography/issues/9207)
    * Operations with encrypted private keys require the
      ``bcrypt`` module, which is installable as ``cryptography[ssh]`` as well.


Configuration
-------------
Peer communication
~~~~~~~~~~~~~~~~~~
To be able to remotely sign certificates, it is required to configure the Salt
master to allow :term:`Peer Communication`:

.. code-block:: yaml

    # /etc/salt/master.d/peer.conf

    peer:
      .*:
        - ssh_pki.sign_remote_certificate

In order for the :term:`Compound Matcher` to work with restricting signing
policies to a subset of minions, in addition calls to :py:func:`match.compound <salt.modules.match.compound>`
by the minion acting as the CA must be permitted:

.. code-block:: yaml

    # /etc/salt/master.d/peer.conf

    peer:
      .*:
        - ssh_pki.sign_remote_certificate

      ca_server:
        - match.compound

Signing policies
~~~~~~~~~~~~~~~~
In addition, the minion representing the CA needs to have at least one
signing policy configured, remote calls not referencing one are always
rejected.

The parameters specified in this signing policy override any
parameters passed from the minion requesting the certificate. It can be
configured in the CA minion's pillar, which takes precedence, or any
location :py:func:`config.get <salt.modules.config.get>` looks up in.
Signing policies are defined under ``ssh_signing_policies``.

Special handling
^^^^^^^^^^^^^^^^
In addition to forcing some arguments to a specific value, signing policies
can also specify default and allowed values.

allowed_critical_options
    A list of critical option names that can be requested to be set under
    this policy. Defaults to all (``["*"]``).

allowed_extensions
    A list of extension names that can be requested to be set under
    this policy. Defaults to all (``["*"]``).

allowed_valid_principals
    A list of principals that can be requested to be set under
    this policy. This is an alias for ``valid_principals`` since
    requesting less permissions is always possible.

default_critical_options
    Defines default critical options that can be overridden by the requester.

default_extensions
    Defines default extensions that can be overridden by the requester.

default_valid_principals
    Defines default principals that can be overridden by the requester.
    Defaults to ``allowed_valid_principals``

critical_options
    Values set here are forced as usual, but per critical option.

extensions
    Values set here are forced as usual, but per extension.

max_ttl
    The maximum TTL that can be requested under this policy.

ttl
    The default TTL that can be overridden by the requester.
    Defaults to ``max_ttl``.

Restricting requesters
^^^^^^^^^^^^^^^^^^^^^^
You can restrict which minions can request a certificate under a configured
signing policy by specifying a matcher in ``minions``. This can be a glob
or compound matcher (for the latter, see the notes above).

.. code-block:: yaml

    ssh_signing_policies:
      www_host:
        - minions: 'www*'
        - signing_private_key: /etc/pki/ssh/ca.key
        - ttl: 30d
        - copypath: /etc/pki/ssh/issued_certs/

.. _sshcert-setup:
"""

import base64
import copy
import logging
import os.path
from datetime import datetime, timedelta, timezone

try:
    from cryptography.hazmat.primitives import hashes, serialization

    import salt.utils.sshpki as sshpki
    import salt.utils.x509 as x509util

    HAS_CRYPTOGRAPHY = True
except ImportError:
    HAS_CRYPTOGRAPHY = False

import salt.utils.atomicfile
import salt.utils.files
import salt.utils.functools
import salt.utils.timeutil as time
from salt.exceptions import CommandExecutionError, SaltInvocationError

__virtualname__ = "ssh_pki"

log = logging.getLogger(__name__)


def __virtual__():
    if not HAS_CRYPTOGRAPHY:
        return (False, "Could not load cryptography")
    return __virtualname__


def create_certificate(
    ca_server=None,
    signing_policy=None,
    path=None,
    overwrite=False,
    raw=False,
    **kwargs,
):
    """
    Create an OpenSSH certificate and return an encoded version of it.

    .. note::

        All parameters that take a public key or private key
        can be specified either as a string or a path to a
        local file encoded for OpenSSH.

    CLI Example:

    .. code-block:: bash

        salt '*' ssh_pki.create_certificate private_key=/root/.ssh/id_rsa signing_private_key='/etc/pki/ssh/myca.key'

    ca_server
        Request a remotely signed certificate from another minion acting as
        a CA server. For this to work, a ``signing_policy`` must be specified,
        and that same policy must be configured on the ca_server. See `Signing policies`_
        for details. Also, the Salt master must permit peers to call the
        ``sign_remote_certificate`` function, see `Peer communication`_.

    signing_policy
        The name of a configured signing policy. Parameters specified in there
        are hardcoded and cannot be overridden. This is required for remote signing,
        otherwise optional. See `Signing policies`_ for details.

    copypath
        Create a copy of the issued certificate in this directory.
        The file will be named ``<serial_number>.crt``.

    path
        Instead of returning the certificate, write it to this file path.

    overwrite
        If ``path`` is specified and the file exists, do not overwrite it.
        Defaults to false.

    raw
        Return the encoded raw bytes instead of a string. Defaults to false.

    cert_type
        The certificate type to generate. Either ``user`` or ``host``.
        Required if not specified in the signing policy.

    private_key
        The private key corresponding to the public key the certificate should
        be issued for. Either this or ``public_key`` is required.

    private_key_passphrase
        If ``private_key`` is specified and encrypted, the passphrase to decrypt it.

    public_key
        The public key the certificate should be issued for. Either this or
        ``private_key`` is required.

    signing_private_key
        The private key of the CA that should be used to sign the certificate. Required.

    signing_private_key_passphrase
        If ``signing_private_key`` is encrypted, the passphrase to decrypt it.

    serial_number
        A serial number to be embedded in the certificate. If unspecified, will
        autogenerate one. This should be an integer, either in decimal or
        hexadecimal notation.

    not_before
        Set a specific date the certificate should not be valid before.
        The format should follow ``%Y-%m-%d %H:%M:%S`` and will be interpreted as GMT/UTC.
        Defaults to the time of issuance.

    not_after
        Set a specific date the certificate should not be valid after.
        The format should follow ``%Y-%m-%d %H:%M:%S`` and will be interpreted as GMT/UTC.
        If unspecified, defaults to the current time plus ``ttl``.

    ttl
        If ``not_after`` is unspecified, a time string (like ``30d`` or ``12h``)
        or the number of seconds from the time of issuance the certificate
        should be valid for. Defaults to ``30d`` for host certificates
        and ``24h`` for client certificates.

    critical_options
        A mapping of critical option name to option value to set on the certificate.
        If an option does not take a value, specify it as ``true``.

        Example:

        .. code-block:: bash

            salt-call ssh_pki.create_certificate [...] \
              critical_options='{"force-command": "/usr/bin/id", "verify-required": true}'

    extensions
        A mapping of extension name to extension value to set on the certificate.
        If an extension does not take a value, specify it as ``true``.

        Example:

        .. code-block:: bash

            salt-call ssh_pki.create_certificate [...] \
              extensions='{"custom-option@my.org": "foobar", "permit-pty": true}'

    valid_principals
        A list of valid principals.

    all_principals
        Allow any principals. Defaults to false.

    key_id
        Specify a string-valued key ID for the signed public key.
        When the certificate is used for authentication, this value will be
        logged in plaintext.
    """
    kwargs = {k: v for k, v in kwargs.items() if not k.startswith("_")}

    if not kwargs.get("signing_private_key") and not ca_server:
        raise SaltInvocationError(
            "Creating a certificate locally at least requires a signing private key."
        )

    if path and not overwrite and __salt__["file.file_exists"](path):
        raise CommandExecutionError(
            f"The file at {path} exists and overwrite was set to false"
        )
    if ca_server:
        if signing_policy is None:
            raise SaltInvocationError(
                "signing_policy must be specified to request a certificate from "
                "a remote ca_server"
            )
        cert = _create_certificate_remote(ca_server, signing_policy, **kwargs)
    else:
        sshpki.merge_signing_policy(_get_signing_policy(signing_policy), kwargs)
        cert = _create_certificate_local(**kwargs)

    out = cert.public_bytes()

    if path is None:
        if raw:
            return out
        return out.decode()

    with salt.utils.files.fopen(path, "wb") as fp_:
        fp_.write(out)
    return f"Certificate written to {path}"


create_certificate_ssh = salt.utils.functools.alias_function(
    create_certificate, "create_certificate_ssh"
)


def _create_certificate_remote(
    ca_server, signing_policy, private_key=None, private_key_passphrase=None, **kwargs
):
    private_key_loaded = None
    if private_key:
        private_key_loaded = sshpki.load_privkey(
            private_key, passphrase=private_key_passphrase
        )
        kwargs["public_key"] = sshpki.encode_public_key(private_key_loaded.public_key())
    elif kwargs.get("public_key"):
        kwargs["public_key"] = sshpki.encode_public_key(
            sshpki.load_pubkey(kwargs["public_key"])
        )

    result = _query_remote(ca_server, signing_policy, kwargs)
    try:
        return sshpki.load_cert(result)
    except (CommandExecutionError, SaltInvocationError) as err:
        raise CommandExecutionError(
            f"ca_server did not return a certificate: {result}"
        ) from err


def _create_certificate_local(copypath=None, **kwargs):
    builder, signing_private_key = sshpki.build_crt(**kwargs)
    cert = builder.sign(signing_private_key)

    if copypath:
        with salt.utils.files.fopen(
            os.path.join(copypath, f"{cert.serial:x}.crt"), "wb"
        ) as fp_:
            fp_.write(cert.public_bytes())
    return cert


def create_private_key(
    algo="rsa",
    keysize=None,
    passphrase=None,
    path=None,
    pubkey_suffix=".pub",
    overwrite=False,
    raw=False,
):
    """
    Create a private key.

    CLI Example:

    .. code-block:: bash

        salt '*' ssh_pki.create_private_key algo=ec keysize=384

    algo
        The digital signature scheme the private key should be based on.
        Available: ``rsa``, ``ec``, ``ed25519``. Defaults to ``rsa``.

    keysize
        For ``rsa``, specifies the bitlength of the private key (2048, 3072, 4096).
        For ``ec``, specifies the NIST curve to use (256, 384, 521).
        Irrelevant for ``ed25519``.
        Defaults to 3072 for RSA and 256 for EC.

    passphrase
        If this is specified, the private key will be encrypted using this
        passphrase. The encryption algorithm cannot be selected, it will be
        determined automatically as the best available one.

    path
        Instead of returning the private key, write it to this file path.
        The file will be written with ``0600`` permissions if it does not exist.

    pubkey_suffix
        If ``path`` is specified, write the corresponding pubkey to the same path
        as the private key with this suffix. Set this to false to skip
        writing the public key. Defaults to ``.pub``.

    overwrite
        If ``path`` is specified and the file exists, overwrite it.
        Defaults to false.

    raw
        Return the encoded raw bytes instead of a string. Defaults to false.
    """
    if algo == "rsa" and keysize is None:
        keysize = 3072

    if path and not overwrite and __salt__["file.file_exists"](path):
        raise CommandExecutionError(
            f"The file at {path} exists and overwrite was set to false"
        )

    priv = encode_private_key(
        _generate_pk(algo=algo, keysize=keysize),
        passphrase=passphrase,
    )
    pub = get_public_key(priv, passphrase=passphrase)

    if path is None:
        if raw:
            return priv.encode()
        return {
            "private_key": priv,
            "public_key": pub,
        }

    salt.utils.atomicfile.safe_atomic_write(path, priv)
    if pubkey_suffix:
        salt.utils.atomicfile.safe_atomic_write(path + pubkey_suffix, pub)

    return f"Private key written to {path}"


def encode_private_key(
    private_key,
    private_key_passphrase=None,
    passphrase=None,
    raw=False,
):
    """
    Create an encoded representation of a private key.

    CLI Example:

    .. code-block:: bash

        salt '*' ssh_ppki.encode_private_key /root/.ssh/id_rsa passphrase=hunter1

    private_key
        The private key to encode.

    private_key_passphrase
        The passphrase that protects the private key.
        Leave unspecified if there is none currently.

    passphrase
        If this is specified, the private key will be encrypted using this
        passphrase. The encryption algorithm cannot be selected, it will be
        determined automatically as the best available one.

    raw
        Return the encoded raw bytes instead of a string. Defaults to false.
    """
    private_key = sshpki.load_privkey(private_key, passphrase=private_key_passphrase)
    if passphrase is None:
        cipher = serialization.NoEncryption()
    else:
        if isinstance(passphrase, str):
            passphrase = passphrase.encode()
        cipher = serialization.BestAvailableEncryption(passphrase)

    pk_bytes = private_key.private_bytes(
        encoding=serialization.Encoding.PEM,
        format=serialization.PrivateFormat.OpenSSH,
        encryption_algorithm=cipher,
    )

    if raw:
        return pk_bytes
    return pk_bytes.decode()


def expires(certificate, ttl=0):
    """
    Determine whether a certificate will expire or has expired already.
    Returns a boolean only.

    CLI Example:

    .. code-block:: bash

        salt '*' ssh_pki.expires /root/.ssh/id_rsa.crt days=7

    certificate
        The certificate to check.

    ttl
        If specified, determine expiration x seconds in the future.
        Can also be specified as a time string like ``30d`` or ``1.5h``.
        Defaults to ``0``, which checks for the current time.
    """
    cert = sshpki.load_cert(certificate)
    # dates are encoded in UTC/GMT, they are returned as a naive datetime object
    return datetime.fromtimestamp(cert.valid_before, tz=timezone.utc) <= datetime.now(
        tz=timezone.utc
    ) + timedelta(seconds=time.timestring_map(ttl))


def get_private_key_size(private_key, passphrase=None):
    """
    Return information about the key size of a private key (RSA/EC).

    CLI Example:

    .. code-block:: bash

        salt '*' ssh_pki.get_private_key_size /root/.ssh/id_rsa

    private_key
        The private key to check.

    passphrase
        If ``private_key`` is encrypted, the passphrase to decrypt it.
    """
    privkey = sshpki.load_privkey(private_key, passphrase=passphrase)
    if not hasattr(privkey, "key_size"):
        # Edwards-curve keys
        return None
    return privkey.key_size


def get_public_key(key, passphrase=None):
    """
    Returns a public key derived from some reference.
    The reference should be a public key, certificate or private key.

    CLI Example:

    .. code-block:: bash

        salt '*' ssh_pki.get_public_key /root/.ssh/id_rsa

    key
        A reference to the structure to look the public key up for.

    passphrase
        If ``key`` is encrypted, the passphrase to decrypt it.
    """
    return sshpki.get_public_key(key, passphrase=passphrase)


def get_signing_policy(signing_policy, ca_server=None):
    """
    Returns the specified named signing policy.

    CLI Example:

    .. code-block:: bash

        salt '*' ssh_pki.get_signing_policy www

    signing_policy
        The name of the signing policy to return.

    ca_server
        If this is set, the CA server will be queried for the
        signing policy instead of looking it up locally.
    """
    if ca_server is None:
        return _get_signing_policy(signing_policy)
    # Cache signing policies from remote during this run
    # to reduce unnecessary resource usage.
    ckey = "_ssh_pki_policies"
    if ckey not in __context__:
        __context__[ckey] = {}
    if ca_server not in __context__[ckey]:
        __context__[ckey][ca_server] = {}
    if signing_policy not in __context__[ckey][ca_server]:
        policy_ = _query_remote(
            ca_server, signing_policy, {}, get_signing_policy_only=True
        )
        __context__[ckey][ca_server][signing_policy] = policy_
    # only hand out copies of the cached policy
    return copy.deepcopy(__context__[ckey][ca_server][signing_policy])


def read_certificate(certificate):
    """
    Returns a dict containing details of a certificate.

    CLI Example:

    .. code-block:: bash

        salt '*' ssh_pki.read_certificate /root/.ssh/id_rsa.crt

    certificate
        The certificate to read.
    """
    cert = sshpki.load_cert(certificate)
    key_type = x509util.get_key_type(cert.public_key(), as_string=True)
    cert_type = None
    if cert.type == serialization.SSHCertificateType.USER:
        cert_type = "user"
    elif cert.type == serialization.SSHCertificateType.HOST:
        cert_type = "host"

    ret = {
        "cert_type": cert_type,
        "valid_principals": (
            "all"
            if cert.valid_principals == []
            else [p.decode() for p in cert.valid_principals]
        ),
        "key_id": cert.key_id.decode(),
        "key_size": cert.public_key().key_size if key_type in ["ec", "rsa"] else None,
        "key_type": key_type,
        "serial_number": x509util.dec2hex(cert.serial),
        "issuer_public_key": sshpki.encode_public_key(cert.signature_key()).decode(),
        "not_before": datetime.fromtimestamp(
            cert.valid_after, tz=timezone.utc
        ).strftime(x509util.TIME_FMT),
        "not_after": datetime.fromtimestamp(
            cert.valid_before, tz=timezone.utc
        ).strftime(x509util.TIME_FMT),
        "public_key": sshpki.encode_public_key(cert.public_key()).decode(),
        "critical_options": _parse_options(cert),
        "extensions": _parse_extensions(cert),
    }

    for pubkey in ("issuer_public_key", "public_key"):
        pubkey_raw = base64.b64decode(ret[pubkey].split(" ")[1])
        pubkey_sha256 = hashes.Hash(hashes.SHA256())
        pubkey_sha256.update(pubkey_raw)
        pubkey_sha256_digest = pubkey_sha256.finalize()
        ret[f"{pubkey}_fingerprints"] = {
            "sha256": base64.b64encode(pubkey_sha256_digest).decode()
        }

        if __opts__["fips_mode"] is False:
            pubkey_md5 = hashes.Hash(hashes.MD5())
            pubkey_md5.update(pubkey_raw)
            pubkey_md5_fingerprint = pubkey_md5.finalize().hex()  # nosec
            ret[f"{pubkey}_fingerprints"]["md5"] = x509util.pretty_hex(
                pubkey_md5_fingerprint
            ).lower()
    return ret


def sign_remote_certificate(
    signing_policy, kwargs, get_signing_policy_only=False, **more_kwargs
):
    """
    Request a certificate to be remotely signed according to a signing policy.
    This is mostly for internal use and does not make much sense on the CLI.

    CLI Example:

    .. code-block:: bash

        salt '*' ssh_pki.sign_remote_certificate www kwargs="{'public_key': '/etc/pki/ssh/www.key'}"

    signing_policy
        The name of the signing policy to use. Required.

    kwargs
        A dict containing all the arguments to be passed into the
        :py:func:`ssh_pki.create_certificate <salt.modules.x509_v2.create_certificate>` function.

    get_signing_policy_only
        Only return the named signing policy. Defaults to false.
    """
    ret = {"data": None, "errors": []}
    try:
        signing_policy = _get_signing_policy(signing_policy)
        if not signing_policy:
            ret["errors"].append(
                "signing_policy must be specified and defined on signing minion"
            )
            return ret
        if "minions" in signing_policy:
            if "__pub_id" not in more_kwargs:
                ret["errors"].append(
                    "minion sending this request could not be identified"
                )
                return ret
            # also pop "minions" to avoid leaking more details than necessary
            if not _match_minions(
                signing_policy.pop("minions"), more_kwargs["__pub_id"]
            ):
                ret["errors"].append(
                    "minion not permitted to use specified signing policy"
                )
                return ret

        if get_signing_policy_only:
            # This is relevant for the state module to be able to check for changes
            # without generating and signing a new certificate every time.
            # remove unnecessary sensitive information
            signing_private_key = signing_policy.pop("signing_private_key", None)
            signing_private_key_passphrase = signing_policy.pop(
                "signing_private_key_passphrase", None
            )
            # ensure to deliver the signing public key as well, not a file path
            if signing_private_key is not None:
                try:
                    signing_private_key = sshpki.load_privkey(
                        signing_private_key, passphrase=signing_private_key_passphrase
                    )
                except (CommandExecutionError, SaltInvocationError) as err:
                    ret["data"] = None
                    ret["errors"].append(str(err))
                    return ret
                signing_policy["signing_public_key"] = sshpki.encode_public_key(
                    signing_private_key.public_key()
                )
            ret["data"] = signing_policy
            return ret
        sshpki.merge_signing_policy(signing_policy, kwargs)
        # ensure the certificate will be issued from this minion
        kwargs.pop("ca_server", None)
    except Exception as err:  # pylint: disable=broad-except
        log.error(str(err))
        return {
            "data": None,
            "errors": [
                "Failed building the signing policy. See CA server log for details."
            ],
        }
    try:
        cert = _create_certificate_local(**kwargs)
        ret["data"] = cert.public_bytes()
        return ret
    except Exception as err:  # pylint: disable=broad-except
        ret["data"] = None
        ret["errors"].append(str(err))
        return ret

    err_message = "Internal error. This is most likely a bug."
    log.error(err_message)
    return {"data": None, "errors": [err_message]}


def _query_remote(ca_server, signing_policy, kwargs, get_signing_policy_only=False):
    result = __salt__["publish.publish"](
        ca_server,
        "ssh_pki.sign_remote_certificate",
        arg=[signing_policy, kwargs, get_signing_policy_only],
    )

    if not result:
        raise SaltInvocationError(
            "ca_server did not respond."
            " Salt master must permit peers to"
            " call the sign_remote_certificate function."
        )
    result = result[next(iter(result))]
    if not isinstance(result, dict) or "data" not in result:
        log.error(f"Received invalid return value from ca_server: {result}")
        raise CommandExecutionError(
            "Received invalid return value from ca_server. See minion log for details"
        )
    if result.get("errors"):
        raise CommandExecutionError(
            "ca_server reported errors:\n" + "\n".join(result["errors"])
        )
    return result["data"]


def verify_private_key(private_key, public_key, passphrase=None):
    """
    Verify that a private key belongs to the specified public key.

    CLI Example:

    .. code-block:: bash

        salt '*' ssh_pki.verify_private_key /root/.ssh/id_rsa /root/.ssh/id_rsa.crt

    private_key
        The private key to check.

    public_key
        The certificate (or any reference that can be passed
        to ``get_public_key``) to retrieve the public key from.

    passphrase
        If ``private_key`` is encrypted, the passphrase to decrypt it.
    """
    privkey = sshpki.load_privkey(private_key, passphrase=passphrase)
    pubkey = sshpki.load_pubkey(get_public_key(public_key))
    return x509util.is_pair(pubkey, privkey)


def verify_signature(certificate, signing_pub_key, signing_pub_key_passphrase=None):
    """
    Verify that a signature on a certificate was made
    by the private key corresponding to the public key associated
    with the specified certificate.

    CLI Example:

    .. code-block:: bash

        salt '*' ssh_pki.verify_signature /root/.ssh/id_rsa.crt /etc/pki/ssh/myca.pub

    certificate
        The certificate to check the signature on.

    signing_pub_key
        Any reference that can be passed to ``get_public_key`` to retrieve
        the public key of the signing entity from.

    signing_pub_key_passphrase
        If ``signing_pub_key`` is encrypted, the passphrase to decrypt it.
    """
    return sshpki.verify_signature(
        certificate,
        signing_pub_key,
        signing_pub_key_passphrase=signing_pub_key_passphrase,
    )


def _generate_pk(algo="rsa", keysize=None):
    if algo == "rsa":
        return x509util.generate_rsa_privkey(keysize=keysize or 2048)
    if algo == "ec":
        return x509util.generate_ec_privkey(keysize=keysize or 256)
    if algo == "ed25519":
        return x509util.generate_ed25519_privkey()
    raise SaltInvocationError(
        f"Invalid algorithm specified for generating private key: {algo}. Valid: "
        "rsa, ec, ed25519, ed448"
    )


def _get_signing_policy(name):
    if name is None:
        return {}
    policies = __salt__["pillar.get"]("ssh_signing_policies", {}).get(name)
    policies = policies or __salt__["config.get"]("ssh_signing_policies", {}).get(name)
    return policies or {}


def _parse_extensions(cert):
    ret = {}
    for ext, val in cert.extensions.items():
        try:
            val = val.decode()
            ret[ext.decode()] = val if val != "" else True
        except UnicodeDecodeError as err:
            log.warning(f"Failed decoding extension: {err}")
    return ret


def _parse_options(cert):
    ret = {}
    for ext, val in cert.critical_options.items():
        try:
            val = val.decode()
            ret[ext.decode()] = val if val != "" else True
        except UnicodeDecodeError as err:
            log.warning(f"Failed decoding option: {err}")
    return ret


def _match_minions(test, minion):
    if "@" in test:
        # Ask the master if the requesting minion matches a compound expression.
        match = __salt__["publish.runner"]("match.compound_matches", arg=[test, minion])
        if match is None:
            raise CommandExecutionError(
                "Could not check minion match for compound expression. "
                "Is this minion allowed to run `match.compound_matches` on the master?"
            )
        try:
            return match["res"] == minion
        except (KeyError, TypeError) as err:
            raise CommandExecutionError(
                "Invalid return value of match.compound_matches."
            ) from err
        # The following line should never be reached.
        return False
    return __salt__["match.glob"](test, minion)



"""
Fixtures for TCP transport functional tests.
"""

# pylint: disable=unused-import

import pytest

import salt.utils.process

# Import SSL certificate fixtures  # noqa: F401
from tests.support.pytest.transport_ssl import ssl_ca_cert_key  # noqa: F401
from tests.support.pytest.transport_ssl import ssl_client_cert_key  # noqa: F401
from tests.support.pytest.transport_ssl import ssl_invalid_ca_cert_key  # noqa: F401
from tests.support.pytest.transport_ssl import ssl_invalid_client_cert_key  # noqa: F401
from tests.support.pytest.transport_ssl import ssl_invalid_server_cert_key  # noqa: F401
from tests.support.pytest.transport_ssl import ssl_master_config  # noqa: F401
from tests.support.pytest.transport_ssl import ssl_minion_config  # noqa: F401
from tests.support.pytest.transport_ssl import ssl_minion_config_no_cert  # noqa: F401
from tests.support.pytest.transport_ssl import (  # noqa: F401; noqa: F401; noqa: F401
    ssl_master_config_invalid_cert,
    ssl_minion_config_invalid_cert,
    ssl_server_cert_key,
)


@pytest.fixture
def process_manager():
    """Process manager for transport tests."""
    pm = salt.utils.process.ProcessManager()
    try:
        yield pm
    finally:
        pm.terminate()



import io
import sys
import tarfile

import pytest

import salt.utils.tarfileutil as tfutil
from tests.support.mock import patch


def _minimal_tar_bytes(name: str = "hello.txt", data: bytes = b"hi") -> bytes:
    buf = io.BytesIO()
    with tarfile.open(fileobj=buf, mode="w") as tf:
        ti = tarfile.TarInfo(name=name)
        ti.size = len(data)
        tf.addfile(ti, io.BytesIO(data))
    return buf.getvalue()


def test_extractall_writes_member(tmp_path):
    data = _minimal_tar_bytes()
    with tarfile.open(fileobj=io.BytesIO(data), mode="r") as tar:
        tfutil.extractall(tar, str(tmp_path))  # nosec B202
    assert (tmp_path / "hello.txt").read_bytes() == b"hi"


def test_extract_writes_member(tmp_path):
    data = _minimal_tar_bytes()
    with tarfile.open(fileobj=io.BytesIO(data), mode="r") as tar:
        member = tar.getmember("hello.txt")
        tfutil.extract(tar, member, str(tmp_path))
    assert (tmp_path / "hello.txt").read_bytes() == b"hi"


@pytest.mark.skipif(
    sys.version_info < (3, 12), reason="filter kw only enforced on 3.12+"
)
def test_extractall_passes_data_filter_on_py312(tmp_path):
    data = _minimal_tar_bytes()
    tar = tarfile.open(fileobj=io.BytesIO(data), mode="r")
    try:
        with patch.object(tar, "extractall", wraps=tar.extractall) as m:
            tfutil.extractall(tar, str(tmp_path))  # nosec B202
        m.assert_called_once()
        _, kwargs = m.call_args
        assert kwargs.get("filter") == "data"
    finally:
        tar.close()


@pytest.mark.skipif(
    sys.version_info < (3, 12), reason="filter kw only enforced on 3.12+"
)
def test_extract_passes_data_filter_on_py312(tmp_path):
    data = _minimal_tar_bytes()
    tar = tarfile.open(fileobj=io.BytesIO(data), mode="r")
    try:
        member = tar.getmember("hello.txt")
        with patch.object(tar, "extract", wraps=tar.extract) as m:
            tfutil.extract(tar, member, str(tmp_path))
        m.assert_called_once()
        _, kwargs = m.call_args
        assert kwargs.get("filter") == "data"
    finally:
        tar.close()
